# -*- coding: utf-8 -*-


# TODO: login using SSL certificate (like in koji)


"""
ClientCommandContainer HOWTO
============================
# ClientCommandContainer extends kobo.cli.CommandContainer
# by adding 'hub' attribute and 'set_hub' method.


1) Import modules
-----------------
import kobo.client
import kobo.client.commands
# Assuming all commands are in <project_name>/commands/cmd_*.py modules.
import <project_name>.commands


2) Inherit the container
------------------------
# Inherit container to make sure nobody will change plugins I registered.
class <ProjectName>CommandContainer(kobo.client.ClientCommandContainer):
    pass


2) Register plugins
-------------------
<ProjectName>CommandContainer.register_module(kobo.client.commands, prefix="cmd_")
<ProjectName>CommandContainer.register_module(<project_name>.commands, prefix="cmd_")


3) Define and call main() function
----------------------------------
def main(args=None):
    config_file = os.environ.get("<PROJECT_NAME>_CONFIG_FILE", "/etc/<project_name>.conf")
    conf = kobo.conf.PyConfigParser()
    conf.load_from_file(config_file)
    command_container = <ProjectName>CommandContainer(conf)
    parser = kobo.cli.CommandOptionParser(command_container=command_container, add_username_password_options=True)
    parser.run(args)
    sys.exit(0)


4) Create commands
------------------
# Commands should be placed in a separate module, usually <project_name>.commands.
# One command per sub-module makes things readable.

import kobo.client

class Add_User(kobo.client.ClientCommand):
    '''add a user account'''
    enabled = True
    admin = True

    def options(self):
        self.parser.usage = "%%prog %s" % self.normalized_name

    def run(self, *args, **kwargs):
        username = kwargs.pop("username", None)
        password = kwargs.pop("password", None)

        self.set_hub(username, password)
        # self.hub.client.add_user(...)
"""


import os
import base64
import hashlib
import ssl
import six.moves.urllib.parse as urlparse
from six.moves import xmlrpc_client as xmlrpclib

import kobo.conf
import kobo.cli
import kobo.http
import kobo.xmlrpc
from kobo.exceptions import AuthenticationError, ImproperlyConfigured


__all__ = (
    "CommandContainer",
    "CommandOptionParser",
    "ClientCommand",
    "ClientCommandContainer",
    "HubProxy",
    "Option",
)


class ClientCommandContainer(kobo.cli.CommandContainer):

    def __init__(self, conf, **kwargs):
        self.conf = kobo.conf.PyConfigParser()
        self.conf.load_from_conf(conf)
        self.conf.load_from_dict(kwargs)

    def set_hub(self, username=None, password=None):
        if username:
            if password is None:
                password = kobo.cli.password_prompt(default_value=password)
            self.conf["AUTH_METHOD"] = "password"
            self.conf["USERNAME"] = username
            self.conf["PASSWORD"] = password

        self.hub = HubProxy(conf=self.conf)


class ClientCommand(kobo.cli.Command):
    pass


class HubProxy(object):
    """A Hub client (thin ServerProxy wrapper)."""

    def __init__(self, conf, client_type=None, logger=None, transport=None, auto_logout=True, **kwargs):
        self._conf = kobo.conf.PyConfigParser()
        self._hub = None

        # load default config
        default_config = os.path.abspath(os.path.join(os.path.dirname(__file__), "default.conf"))
        self._conf.load_from_file(default_config)

        # update config with another one
        if conf is not None:
            self._conf.load_from_conf(conf)

        # update config with kwargs
        self._conf.load_from_dict(kwargs)

        # initialize properties
        self._client_type = client_type or "client"
        self._hub_url = self._conf["HUB_URL"]
        self._auth_method = self._conf["AUTH_METHOD"]
        self._auto_logout = auto_logout
        self._logger = logger
        self._logged_in = False

        if transport is not None:
            self._transport = transport
        else:
            transport_args = {}
            if self._hub_url.startswith("https://"):
                TransportClass = kobo.xmlrpc.retry_request_decorator(kobo.xmlrpc.SafeCookieTransport)
                if hasattr(ssl, 'create_default_context'):
                    ssl_context = ssl.create_default_context()
                    if self._conf.get('CA_CERT'):
                        ssl_context.load_verify_locations(cafile=self._conf['CA_CERT'])
                    else:
                        ssl_context.check_hostname = False
                        ssl_context.verify_mode = ssl.CERT_NONE
                    transport_args['context'] = ssl_context
            else:
                TransportClass = kobo.xmlrpc.retry_request_decorator(kobo.xmlrpc.CookieTransport)
            self._transport = TransportClass(**transport_args)

        # self._hub is created here
        try:
            self._login(verbose=self._conf.get("DEBUG_XMLRPC"))
        except KeyboardInterrupt:
            raise
        except Exception as ex:
            self._logger and self._logger.warn("Authentication failed")
            raise

    def __del__(self):
        if hasattr(self._transport, "retry_count"):
            self._transport.retry_count = 0
        if getattr(self, "_auto_logout", False) and self._logged_in:
            try:
                self._logout()
            except:
                pass

    def __getattr__(self, name):
        try:
            return getattr(self._hub, name)
        except:
            raise AttributeError("'%s' object has no attribute '%s'" % (self.__class__.__name__, name))

    def _login(self, force=False, verbose=False):
        """Login to the hub.
        - self._hub instance is created in this method
        - session information is stored in a cookie in self._transport
        """

        login_method_name = "_login_%s" % self._auth_method
        if not hasattr(self, login_method_name):
            raise ImproperlyConfigured("Unknown authentication method: %s" % self._auth_method)

        # create new self._hub instance (only once, when calling constructor)
        if self._hub is None:
            self._hub = xmlrpclib.ServerProxy("%s/%s/" % (self._hub_url, self._client_type), allow_none=True, transport=self._transport, verbose=verbose)

        if force or self._hub.auth.renew_session():
            self._logger and self._logger.info("Creating new session...")
            try:
                # logout to delete current session information
                self._logout()
            except KeyboardInterrupt:
                raise
            except Exception as ex:
                self._logger and self._logger.error("Failed to log out: %s" % ex)

            try:
                login_method = getattr(self, login_method_name)
                login_method()
                self._logged_in = True
            except KeyboardInterrupt:
                raise
            except Exception as ex:
                self._logger and self._logger.debug("Failed to create new session: %s" % ex)
            else:
                self._logger and self._logger.info("New session created.")

    def _logout(self):
        """Logout from hub"""
        if hasattr(self, "_hub"):
            self._hub.auth.logout()

    def _login_password(self):
        """Login using username and password."""
        username = self._conf.get("USERNAME")
        password = self._conf.get("PASSWORD")
        if not username:
            raise AuthenticationError("USERNAME is not set")
        self._hub.auth.login_password(username, password)

    def _login_worker_key(self):
        """Login using worker key."""
        worker_key = self._conf.get("WORKER_KEY")
        if not worker_key:
            raise AuthenticationError("WORKER_KEY is not set")
        self._hub.auth.login_worker_key(worker_key)

    def _login_krbv(self):
        """Login using kerberos credentials (uses python-krbV)."""

        def get_server_principal(service=None, realm=None):
            """Convert hub url to kerberos principal."""
            hostname = urlparse.urlparse(self._hub_url)[1]
            # remove port from hostname
            hostname = hostname.split(":")[0]

            if realm is None:
                # guess realm: last two parts from hostname
                realm = ".".join(hostname.split(".")[-2:]).upper()
            if service is None:
                service = "HTTP"
            return '%s/%s@%s' % (service, hostname, realm)


        # read default values from settings
        principal = self._conf.get("KRB_PRINCIPAL")
        keytab = self._conf.get("KRB_KEYTAB")
        service = self._conf.get("KRB_SERVICE")
        realm = self._conf.get("KRB_REALM")
        ccache = self._conf.get("KRB_CCACHE")
        proxyuser = self._conf.get("KRB_PROXYUSER")

        import krbV
        ctx = krbV.default_context()

        if ccache is not None:
            ccache = krbV.CCache(name='FILE:' + ccache, context=ctx)
        else:
            ccache = ctx.default_ccache()

        if principal is not None:
            if keytab is not None:
                cprinc = krbV.Principal(name=principal, context=ctx)
                keytab = krbV.Keytab(name=keytab, context=ctx)
                ccache.init(cprinc)
                ccache.init_creds_keytab(principal=cprinc, keytab=keytab)
            else:
                raise ImproperlyConfigured("Cannot specify a principal without a keytab")
        else:
            # connect using existing credentials
            cprinc = ccache.principal()

        sprinc = krbV.Principal(name=get_server_principal(service=service, realm=realm), context=ctx)

        ac = krbV.AuthContext(context=ctx)
        ac.flags = krbV.KRB5_AUTH_CONTEXT_DO_SEQUENCE | krbV.KRB5_AUTH_CONTEXT_DO_TIME
        ac.rcache = ctx.default_rcache()

        # create and encode the authentication request
        try:
            ac, req = ctx.mk_req(server=sprinc, client=cprinc, auth_context=ac, ccache=ccache, options=krbV.AP_OPTS_MUTUAL_REQUIRED)
        except krbV.Krb5Error as ex:
            if getattr(ex, "err_code", None) == -1765328377:
                ex.message += ". Make sure you correctly set KRB_REALM (current value: %s)." % realm
                ex.args = (ex.err_code, ex.message)
            raise ex
        req_enc = base64.encodestring(req)

        self._hub.auth.login_krbv(req_enc)

    def upload_file(self, file_name, target_dir):
        scheme, netloc, path, params, query, fragment = urlparse.urlparse("%s/upload/" % self._hub_url)
        if ":" in netloc:
            host, port = netloc.split(":", 1)
        else:
            host, port = netloc, None

        sum = hashlib.sha256()
        fo = open(file_name, "rb")
        while True:
            chunk = fo.read(1024 ** 2)
            if not chunk:
                break
            sum.update(chunk)
        fo.close()
        checksum = sum.hexdigest().lower()

        fsize = os.path.getsize(file_name)
        upload_id, upload_key = self.upload.register_upload(os.path.basename(file_name), checksum, fsize, target_dir)

        secure = (scheme == "https")
        upload = kobo.http.POSTTransport()
        upload.add_variable("upload_id", upload_id)
        upload.add_variable("upload_key", upload_key)
        upload.add_file("file", file_name)

        err_code, err_msg = upload.send_to_host(host, path, port, secure)
        return upload_id, err_code, err_msg

    def upload_task_log(self, file_obj, task_id, remote_file_name, append=True, mode=0o644):
        """
        Upload a task log to the hub.

        @param file_obj: file object (or StringIO, etc.) with the log
        @type  file_obj: file
        @param task_id: task ID
        @type  task_id: int
        @param remove_file_name: relative path on hub to the log file
        @type  remove_file_name: str
        @param append: append at the end of existing file instead of rewriting it
        @type  append: bool
        @param mode: file perms (example: 0644)
        @type  mode: int
        """

        for (chunk_start, chunk_len, chunk_checksum, encoded_chunk) in kobo.xmlrpc.encode_xmlrpc_chunks_iterator(file_obj):
            if append:
                chunk_start = -1
                if chunk_len == -1:
                    # skip finializing chunk
                    break
            self._hub.worker.upload_task_log(task_id, remote_file_name, mode, chunk_start, chunk_len, chunk_checksum, encoded_chunk)


from six.moves.xmlrpc_client import Fault


# default implementation of Fault.__repr__ is:
#    "<Fault %s: %s>" % (self.faultCode, repr(self.faultString))
# repr of string does not escape newlines ('\n') and produces very ugly output
# so using direct string is much nicer for users
def fault_repr(self):
    return "<Fault %s: %s>" % (self.faultCode, str(self.faultString))

Fault.__repr__ = fault_repr



# -*- coding: utf-8 -*-


"""
This is a menu system for django applications.

It creates 'menu' object which provides access
to functions for manipulation with html menu.

To set the menu up, follow steps below:


1) create project/menu.py
-------------------------
from kobo.django.menu import MenuItem, include
menu = (
    MenuItem("MenuItem-1", "/url/path/", absolute_url=True, menu=(
        MenuItem("MenuItem-1.1", "/url/path/1/", absolute_url=True),
        MenuItem("MenuItem-1.2", "/url/path/2/", absolute_url=True),
    )),
    MenuItem("MenuItem-2", "url_label", ("Developers",), ("app.change_data",)),
    include("project.app.menu"),
    MenuItem.include("project.another_app.menu"),
)

# In this example is MenuItem-1 and it's submenu tree accessible for anybody.
# MenuItem-2 is only for users in group Developers with specific permission.
# Instead of specifying complete tree in one file, you can use include()
# command in similar way as it is used in urls.py (see third menu item).
# include() function is also a staticmethod of MenuItem class (see fourth menu item).
#
# If you leave title empty, the menu item will serve as a delimiter. In such a
# case you can also omit the url.

# can be specified only once in project-wide menu
css_active_class = "active_menu"


2) modify settings.py
---------------------
# set the menu root
ROOT_MENUCONF = "project.menu"

# add menu middleware
MIDDLEWARE_CLASSES = (
    ...
    "kobo.django.menu.middleware.MenuMiddleware",
    ...
)

# add a context processor to add 'menu' variable to each request
TEMPLATE_CONTEXT_PROCESSORS = (
    ...
    "kobo.django.menu.context_processors.menu_context_processor",
    ...
)


3) add menu to a html template
------------------------------
# The simplest use in template is just to display menu as {{ menu }}.
# Default format is nested ul list.
# Active menus/submenus will be tagged with css_active_class.
# Delimiters will be rendered as empty list items with class `delimiter`.

# If you need another approach, you can access directly to all menu items
# and display menu parts by yourself:

# display all menu items from main menu
{% for m in menu.items %}
  {{ m.as_a }}
{% endfor %}

# Delimiters rendered with `as_a` return empty string, so are invisible.

# menu.levelX = active menu item on Xth level
# menu.levelX.items = all visible menu items of the ^^^ menu
{% for m in menu.level1.items %}
  {{ m.as_a }}
{% endfor %}

# If you are using Bootstrap and want to include the menu in a navbar, there is
# a method to render the menu in this way. Note however that in this case you
# can only use one level of nesting. The top level items will be rendered
# directly in the navbar, their children will form a dropdown menu and their
# grand children will be ignored.
{{ m.as_bootstrap_navbar_dropdown_menu }}
"""

import six

from six.moves import range

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.encoding import python_2_unicode_compatible
from kobo.django.django_version import django_version_ge
if django_version_ge('1.10.0'):
    from django.urls import reverse
else:
    from django.core.urlresolvers import reverse
from django.utils.safestring import mark_safe

try:
    from django.utils.encoding import smart_unicode as smart_text
except ImportError:
    from django.utils.encoding import smart_text


__all__ = (
    "include",
    "menu",
    "MenuItem",
)


def include(module):
    """helper function to load nested menus"""
    m = __import__(module, {}, {}, [""])
    return m.menu


@python_2_unicode_compatible
class MenuItem(object):
    """basic menu item - every menuitem can have submenu collections of
    these. Only main menu is special instance of Menu class."""

    # bind include() to the class
    include = staticmethod(include)

    def __init__(self, title, url, acl_groups=None, acl_perms=None, absolute_url=False, menu=None):
        self.title = smart_text(title)
        self._url = url
        self._url_is_resolved = absolute_url
        self.absolute_url = absolute_url
        self.acl_groups = acl_groups and set(acl_groups) or set()
        self.acl_perms = acl_perms and set(acl_perms) or set()
        self.main_menu = None
        self.parent_menu = None
        self.alters_data = False

        self.submenu_list = []
        for i in menu or []:
            if type(i) in (tuple, list):
                self.submenu_list.extend(i)
            else:
                self.submenu_list.append(i)

        self.active = False
        self.depth = 0

    def __repr__(self):
        return "<MenuItem>: %s (%s)" % (self.title, self.url)

    def __len__(self):
        return len(self.url)

    def __str__(self):
        if self.title == "":
            return mark_safe(u"<li class='divider'></li>")
        result = ""
        if self.items:
            result = u"<ul>%s</ul>" % u"".join([six.text_type(i) for i in self.items])
        return mark_safe(u"<li>%s%s</li>" % (self.as_a(), result))

    @property
    def url(self):
        if self._url and not self._url_is_resolved:
            self._url = reverse(self._url)
        self._url_is_resolved = True
        return self._url

    def as_a(self):
        if not self.visible:
            return u""

        if self.active and self.main_menu.css_active_class:
            style = u" class='%s'" % self.main_menu.css_active_class
        else:
            style = u""

        if self.url:
            result = u"<a href='%s'%s>%s</a>" % (self.url, style, self.title)
        else:
            result = six.text_type(self.title)

        return mark_safe(result)

    @property
    def items(self):
        return [i for i in self.submenu_list if i.visible]

    @property
    def first_item(self):
        if self.items:
            return self.items[0]
        return None

    @property
    def last_item(self):
        if self.items:
            return self.items[-1]
        return None

    @property
    def inner_items(self):
        return self.items[1:-1]

    def setup_menu_tree(self, mainmenu_obj):
        if mainmenu_obj != self:
            self.main_menu = mainmenu_obj

        actual_depth = self.depth

        if self.submenu_list:
            mainmenu_obj.depth = max(mainmenu_obj.depth, actual_depth + 1)

        for i in self.submenu_list:
            i.parent_menu = self
            i.depth = actual_depth + 1
            i.setup_menu_tree(mainmenu_obj)
            mainmenu_obj.cached_menuitems.append(i)

    def set_active(self, active):
        self.active = active
        if self.parent_menu is not None:
            self.parent_menu.set_active(active)

    @property
    def visible(self):
        # return False if field should be displayed to user
        if self.main_menu.user.is_superuser:
            return True

        if self.acl_groups:
            if self.acl_groups.intersection(self.main_menu.acl_groups):
                return True
            return False

        if self.acl_perms:
            for perm in self.acl_perms:
                if perm not in self.main_menu.acl_perms:
                    self.main_menu.acl_perms[perm] = self.main_menu.user.has_perm(perm)
                if self.main_menu.acl_perms[perm]:
                    return True
            return False

        return True

    def as_li(self):
        """Render menu item as a list item."""
        if not self.title:
            return mark_safe('<li class="divider"></li>')
        return mark_safe('<li>%s</li>' % self.as_a())

    def as_bootstrap_navbar_dropdown_menu(self):
        """
        Render menu item as a list with a possible dropdown. Note that any
        items nested under children of this item are ignored.
        """
        cls = ''
        sub = ''
        link = self.as_a()
        if self.items:
            link = '<a href="%s" class="dropdown-toggle" data-toggle="dropdown">%s <span class="caret"></span></a>' % (self.url, self.title)
            sub = ''.join([i.as_li() for i in self.items])
            sub = '<ul class="dropdown-menu" role="menu">%s</ul>' % sub
            cls = ' class="dropdown"'

        return mark_safe('<li%s>%s%s</li>' % (cls, link, sub))


@python_2_unicode_compatible
class MainMenu(MenuItem):

    def __init__(self, menu, css_active_class=None):
        MenuItem.__init__(self, "ROOT_MENU", "", absolute_url=True, menu=menu)
        self.user = None
        self.path = ""
        self.cached_menuitems = []
        self.css_active_class = css_active_class or ""
        self.active = None # reference to active menu (overrides MenuItem behavior)

        # set main_menu references, compute menu depth
        self.setup_menu_tree(self)

    def __repr__(self):
        return "<MainMenu>: %s" % (self.path)

    def __str__(self):
        """Return menu as printable <ul> list."""
        return mark_safe(u"<ul>%s</ul>" % "".join([six.text_type(i) for i in self.items]))

    def as_bootstrap_navbar_dropdown_menu(self):
        """
        Return menu as a printable <ul> list appropriate for use in Bootstrap
        navbar. Only one level nesting is supported and the nested items are
        rendered as dropdown menus.
        """
        content = ''.join([i.as_bootstrap_navbar_dropdown_menu() for i in self.items])
        return mark_safe('<ul class="nav navbar-nav">%s</ul>' % content)

    def __getattr__(self, name):
        # get specified submenu level in active menu
        if not name.startswith("level"):
            raise AttributeError("'%s' object has no attribute '%s'" % (self.__class__.__name__, name))

        try:
            level = int(name[5:])
        except ValueError:
            raise AttributeError("'%s' object has no attribute '%s'" % (self.__class__.__name__, name))

        if level not in list(range(1, self.depth + 1)):
            raise AttributeError("'%s' object has no attribute '%s'" % (self.__class__.__name__, name))

        if not self.active:
            return None

        if self.active.depth < level:
            raise AttributeError("'%s' object has no attribute '%s'" % (self.__class__.__name__, name))

        menu = self.active
        while menu.depth > level:
            menu = menu.parent_menu
        return menu

    def setup(self, request):
        self.user = request.user
        self.path = request.get_full_path()
        self.acl_groups = set([i.name for i in request.user.groups.all().only("name")])
        self.acl_perms = {}
        self.find_active_menu()
        return self

    def find_active_menu(self):
        if self.active:
            # reset cached active path
            self.active.set_active(False)

        matches = [i for i in self.cached_menuitems if i.visible and i.url and self.path.startswith(i.url)]
        if not matches:
            self.active = None
            return None

        # find the longest menu match
        matches.sort(key=len, reverse=True)
        found = matches[0]
        found.set_active(True)
        self.active = found
        return found


# load menu configuration from project
if not hasattr(settings, "ROOT_MENUCONF"):
    raise ImproperlyConfigured("'ROOT_MENUCONF' is needed in settings to run kobo.django.menu app.")
menu_module = __import__(settings.ROOT_MENUCONF, {}, {}, [""])
menu = MainMenu(menu_module.menu, getattr(menu_module, "css_active_class", ""))



# -*- coding: utf-8 -*-


import os
import re

import kobo.cli
import kobo.admin
import kobo.shortcuts


class Start_Hub(kobo.cli.Command):
    """create a hub project directory structure in the current directory"""
    enabled = True

    def options(self):
        self.parser.usage = "%%prog %s [options] <project_name>" % self.normalized_name

    def run(self, *args, **kwargs):
        if len(args) < 1:
            self.parser.error("Please specify a name of the project.")

        name = args[0].replace("-", "_")
        directory = os.getcwd()

        try:
            kobo.admin.copy_helper(name, directory, "hub")
        except kobo.admin.TemplateError as ex:
            self.parser.error(ex)

        # code from django/core/management/commands/startproject.py
        # Create a random SECRET_KEY hash, and put it in the main settings.
        main_settings_file = os.path.join(directory, name, 'settings.py')
        settings_contents = open(main_settings_file, 'r').read()
        fp = open(main_settings_file, 'w')
        django_alphabet = "abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*(-_=+)"
        secret_key = kobo.shortcuts.random_string(50, alphabet=django_alphabet)
        settings_contents = re.sub(r"(?<=SECRET_KEY = ')'", secret_key + "'", settings_contents)
        fp.write(settings_contents)
        fp.close()



# -*- coding: utf-8 -*-


import os

from django.conf import settings
from django.db import models
from django.contrib.auth.models import User

from kobo.shortcuts import random_string
from kobo.types import Enum
import six


UPLOAD_STATES = Enum(
    "NEW",
    "STARTED",
    "FINISHED",
    "FAILED",
)


class FileUpload(models.Model):
    owner       = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    name        = models.CharField(max_length=255)
    checksum    = models.CharField(max_length=255)
    size        = models.PositiveIntegerField()
    target_dir = models.CharField(max_length=255)
    upload_key  = models.CharField(max_length=255)
    state       = models.PositiveIntegerField(default=0, choices=UPLOAD_STATES.get_mapping())
    dt_created  = models.DateTimeField(auto_now_add=True)
    dt_finished = models.DateTimeField(null=True, blank=True)

    class Meta:
        ordering = ("-dt_created", "name", )
#        unique_together = (
#            ("name", "target_dir")
#        )

    def export(self):
        result = {
            "owner": self.owner_id,
            "name": self.name,
            "checksum": self.checksum,
            "size": self.size,
            "target_dir": self.target_dir,
            "state": self.state,
        }
        return result

    def get_full_path(self):
        return os.path.abspath(os.path.join(self.target_dir, self.name))

    def __unicode__(self):
        return six.text_type(os.path.join(self.target_dir, self.name))

    def save(self, *args, **kwargs):
        if not self.upload_key:
            self.upload_key = random_string(64)
        if self.state == UPLOAD_STATES['FINISHED']:
            if FileUpload.objects.filter(state = UPLOAD_STATES['FINISHED'], name = self.name).exclude(id = self.id).count() != 0:
                # someone created same upload faster
                self.state == UPLOAD_STATES['FAILED']
        super(FileUpload, self).save(*args, **kwargs)

    def delete(self):
        super(FileUpload, self).delete()
        # if file was successfully uploaded it should be removed from
        # filesystem, otherwise it shouldn't be there
        if self.state == UPLOAD_STATES['FINISHED']:
            try:
                os.unlink(self.get_full_path())
            except OSError as ex:
                if ex.errno != 2:
                    raise

            upload_dir = getattr(settings, "UPLOAD_DIR", None)
            if upload_dir is not None:
                upload_dir = os.path.abspath(upload_dir)
                file_dir = os.path.dirname(self.get_full_path())
                while 1:
                    if not file_dir.startswith(upload_dir):
                        break
                    if file_dir == upload_dir:
                        break
                    try:
                        os.rmdir(file_dir)
                    except OSError as ex:
                        break
                    file_dir = os.path.split(file_dir)[0]



# -*- coding: utf-8 -*-

import os
import re
import io
import sys
import signal
import time


__all__ = (
    "daemonize",
    "get_child_pgids",
    "get_proc_stat",
    "get_process_status",
    "is_success",
    "kill_process_group",
    "kill_group",
)


def daemonize(daemon_func, daemon_pid_file=None, daemon_start_dir="/", daemon_out_log="/dev/null", daemon_err_log="/dev/null", *args, **kwargs):
    """Robustly turn into a UNIX daemon, running in daemon_start_dir."""

    if daemon_pid_file and os.path.exists(daemon_pid_file):
        try:
            f = open(daemon_pid_file, "r")
            pid = f.read()
            f.close()
        except:
            pid = None

        if pid:
            try:
                fn = os.path.join("/proc", pid, "cmdline")
                f = open(fn, "r")
                cmdline = f.read()
                f.close()
            except:
                cmdline = None

            if cmdline and cmdline.find(sys.argv[0]) >= 0:
                sys.stderr.write("A proces is still running, pid: %s\n" % pid)
                sys.exit(1)

    # first fork
    try:
        if os.fork() > 0:
            # exit from first parent
            sys.exit(0)
    except OSError as ex:
        sys.stderr.write("fork #1 failed: (%d) %s\n" % (ex.errno, ex.strerror))
        sys.exit(1)

    # decouple from parent environment
    os.setsid()
    try:
        os.chdir(daemon_start_dir)
    except OSError:
        # fallback to "/" (just in case the first chdir fails on insufficient perms or another OSError)
        os.chdir("/")
    os.umask(0)

    # second fork
    try:
        pid = os.fork()
        if pid > 0:
            # write pid to pid_file
            if daemon_pid_file is not None:
                fd = os.open(daemon_pid_file, os.O_WRONLY | os.O_CREAT, 0o644)
                os.write(fd, b"%s" % str(pid).encode())
                os.close(fd)
            # exit from second parent
            sys.exit(0)
    except OSError as ex:
        sys.stderr.write("fork #2 failed: (%d) %s\n" % (ex.errno, ex.strerror))
        sys.exit(1)

    # redirect stdin, stdout and stderr
    stdin = open("/dev/null", "r")
    # Python 3
    try:
        stdout = io.TextIOWrapper(open(daemon_out_log, "ab+", 0), write_through=True)
        stderr = io.TextIOWrapper(open(daemon_err_log, "ab+", 0), write_through=True)
    # Python 2
    except TypeError:
        stdout = open(daemon_out_log, "a+", 0)
        stderr = open(daemon_err_log, "a+", 0)
    os.dup2(stdin.fileno(), sys.stdin.fileno())
    os.dup2(stdout.fileno(), sys.stdout.fileno())
    os.dup2(stderr.fileno(), sys.stderr.fileno())

    # run the daemon loop
    daemon_func(*args, **kwargs)

#    # delete pid file
#    if daemon_pid_file:
#        try:
#            os.remove(daemon_pid_file)
#        except:
#            pass

    sys.exit(0)


def get_process_status(retval, prefix):
    """Return status description after a process has exited."""

    if type(prefix) in (list, tuple):
        prefix = " ".join(prefix)

    if os.WIFSIGNALED(retval):
        return "%s was killed by signal %i" % (prefix, os.WTERMSIG(retval))
    if os.WIFEXITED(retval):
        return "%s exited with status %i" % (prefix, os.WEXITSTATUS(retval))
    return "%s terminated for unknown reasons" % prefix


def is_success(return_code):
    """Return True if return code indicates successful completion (exited with status 0), False otherwise."""
    if os.WIFEXITED(return_code) and os.WEXITSTATUS(return_code) == 0:
        return True
    return False


procstat_re = re.compile(r"^(?P<pid>-?\d+) \((?P<comm>.+)\) (?P<state>\w) (?P<ppid>-?\d+) (?P<pgid>-?\d+)"
                        +r" (?P<sid>-?\d+) (?P<tty_nr>-?\d+) (?P<tty_pgid>-?\d+) (?P<flags>\d+) (?P<minflt>\d+)"
                        +r" (?P<cminflt>\d+) (?P<majflt>\d+) (?P<cmajflt>\d+) (?P<utime>\d+) (?P<stime>\d+)"
                        +r" (?P<cutime>-?\d+) (?P<cstime>-?\d+) (?P<priority>-?\d+) (?P<nice>-?\d+) (?P<num_threads>\d+)"
                        +r" 0 (?P<itrealvalue>-?\d+) (?P<starttime>-?\d+) (?P<vsize>\d+) (?P<rss>-?\d+)"
                        +r" (?P<startcode>\d+) (?P<endcode>\d+) (?P<startstack>\d+) (?P<kstkesp>\d+) (?P<kstkeip>\d+)"
                        +r" (?P<signal>\d+) (?P<blocked>\d+) (?P<sigignore>\d+) (?P<sigcatch>\d+) (?P<wchan>\d+)"
                        +r" (?P<nswap>\d+) (?P<cnswap>\d+) (?P<exit_signal>-?\d+) (?P<processor>-?\d+) (?P<rt_priority>-?\d+)"
                        +r" (?P<policy>-?\d+) ?(?P<blkio_ticks>\d+)? ?(?P<gtime>\d+) ?(?P<cgtime>-?\d+)?.*$")


def get_proc_stat(pid):
    """Get information from /proc/<PID>/stat.

    See man proc for detail (inaccurate).
    Accurate data can be found in kernel sources: fs/proc/array.c
    """

    # read stat file
    procfile = open("/proc/%s/stat" % pid)
    procdata = procfile.read()
    procfile.close()

    # parse data and store into dictionary
    match = procstat_re.match(procdata)
    if match is not None:
        result = match.groupdict()
        for key in result:
            # keep following field as string
            if key in ("comm", ):
                continue

            # covert rest to integer
            if type(result[key]) is str and result[key].isdigit():
                result[key] = int(result[key])

        return result

    raise IOError("Invalid /proc/%s/stat file" % pid)


def kill_process_group(pid, msg=None, sig=signal.SIGTERM, timeout=5, logger=None):
    """Kill process group with signal, keep trying within timeout.

    Return True if successful, False if not.
    """

    success = True
    for pgid in get_child_pgids(pid)[::-1]:
        # iterate in reverse order so processes whose children are killed might have
        # a chance to cleanup before they"re killed
        success &= kill_group(pgid, msg, sig, timeout, logger)
    return success


# TODO: remove or use *msg* argument
def kill_group(pgid, msg=None, sig=signal.SIGTERM, timeout=5, logger=None):
    """Kill the process group with the given process group ID.
    Return True if the group is successfully killed in the given timeout, False otherwise."""

    incr = 1.0
    t = 0.0

    while t < timeout:
        try:
            pid, retval = os.waitpid(-pgid, os.WNOHANG)
            while pid != 0:
                if logger:
                    logger.info(get_process_status(retval, "kill_group: process %i" % pid))
                pid, retval = os.waitpid(-pgid, os.WNOHANG)
        except OSError as ex:
            # means there are no processes in that process group
            if t == 0.0:
                logger and logger.info("kill_group: Process (pgrp %i) exited" % (pgid))
            else:
                logger and logger.info("kill_group: Killed process (pgrp %i)" % (pgid))
            return True
        else:
            logger and logger.info("kill_group: Process (pgrp %i) exists" % (pgid))

        try:
            os.killpg(pgid, sig)
        except OSError as ex:
            # shouldn't happen
            logger and logger.error("kill_group: Process (pgrp %i): %s" % (pgid, ex))
            continue
        else:
            logger and logger.info("kill_group: Sent signal %i to process (pgrp %i)" % (sig, pgid))

        if t == 0.0:
            time.sleep(0.1)
        else:
            time.sleep(incr)
            t += incr
#        time.sleep(incr)
        t += incr
    logger and logger.error("kill_group: Failed to kill process (pgrp %i)" % (pgid))
    return False


def get_child_pgids(pid):
    """
    Recursively get the children of the process with the given ID.
    Return a list containing the process group IDs of the children
    in depth-first order, without duplicates.
    """

    stats_by_ppid = {}
    pgids = []

    for procdir in os.listdir("/proc"):
        if not procdir.isdigit():
            continue

        try:
            stat = get_proc_stat(procdir)
            stats_by_ppid.setdefault(stat["ppid"], [])
            stats_by_ppid[stat["ppid"]].append(stat)
            if stat["pid"] == pid:
                # put the pgid of the top-level process into the list
                pgids.append(stat["pgid"])
        except (IOError, OSError):
            # We expect IOErrors, because files in /proc may disappear between the listdir() and read().
            # Nothing we can do about it, just move on.
            continue

    if not pgids:
        # assume the pid and pgid of the forked process are the same
        pgids.append(pid)

    pids = [pid]
    while pids:
        for ppid in pids[:]:
            for stat in stats_by_ppid.get(ppid, []):
                # get the /proc entries with ppid as their parent, and append their pgid to the list,
                # then recheck for their children
                if stat["pgid"] not in pgids:
                    pgids.append(stat["pgid"])
                pids.append(stat["pid"])
            pids.remove(ppid)

    return pgids



# -*- coding: utf-8 -*-


import sys

import kobo.cli


__all__ = (
    "main",
)


# register default command plugins
#import kobo.client.commands
#CommandContainer.register_module(kobo.client.commands, prefix="cmd_")


def main():
    command_container = kobo.cli.CommandContainer()
    parser = kobo.cli.CommandOptionParser(command_container=command_container, add_username_password_options=True)
    parser.run()
    sys.exit(0)



# -*- coding: utf-8 -*-

import django

# Only for Django >= 1.7
if 'setup' in dir(django):
    # This has to happen before below imports because they have a hard requirement
    # on settings being loaded before import.
    django.setup()

import unittest2 as unittest

from mock import Mock, PropertyMock, patch

from kobo.hub import middleware


class DummyRequest(object):
    pass


class DummyWorker(object):

    def __init__(self, name=None):
        self.name = name


class TestGetWorker(unittest.TestCase):

    def test_get_worker(self):
        with patch('kobo.hub.middleware.Worker') as worker_mock:
            worker_mock.objects.get.return_value = DummyWorker()
            req = PropertyMock(user=PropertyMock(username='foo/bar'))
            worker = middleware.get_worker(req)
            self.assertIsInstance(worker, DummyWorker)
            worker_mock.objects.get.assert_called_once_with(name='bar')

    def test_get_worker_missing_hostname(self):
        req = PropertyMock(user=PropertyMock(username='username'))
        worker = middleware.get_worker(req)
        self.assertIsNone(worker)

    def test_get_worker_catch_exceptions(self):
        req = PropertyMock(user=Mock(side_effect=ValueError))
        worker = middleware.get_worker(req)
        self.assertIsNone(worker)


class TestLazyWorker(unittest.TestCase):

    def test_lazy_worker_set_cache_variable_if_not_set(self):
        with patch('kobo.hub.middleware.get_worker', return_value=DummyWorker()) as get_worker_mock:
            req = PropertyMock(
                spec=['user'],
                user=PropertyMock(username='foo/bar'),
            )

            cached_worker = middleware.LazyWorker().__get__(req)
            self.assertIsInstance(cached_worker, DummyWorker)

            get_worker_mock.assert_called_once_with(req)

    def test_lazy_worker_do_not_set_cache_variable_if_already_set(self):
        with patch('kobo.hub.middleware.get_worker', return_value=DummyWorker('new-worker')) as get_worker_mock:
            req = PropertyMock(
                spec=['user', '_cached_worker'],
                user=PropertyMock(username='foo/bar'),
                _cached_worker=DummyWorker('cached-worker'),
            )

            cached_worker = middleware.LazyWorker().__get__(req)
            self.assertIsInstance(cached_worker, DummyWorker)
            self.assertEquals(cached_worker.name, 'cached-worker')

            get_worker_mock.assert_not_called()


class TestWorkerMiddleware(unittest.TestCase):

    def test_process_request(self):
        with patch('kobo.hub.middleware.get_worker', return_value=DummyWorker()) as get_worker_mock:
            req = DummyRequest()
            req.user = PropertyMock(username='foo/bar')

            middleware.WorkerMiddleware().process_request(req)
            self.assertIsInstance(req.worker, DummyWorker)
            get_worker_mock.assert_called_once_with(req)

    def test_process_request_missing_user(self):
        req = DummyRequest()

        with self.assertRaises(AssertionError):
            middleware.WorkerMiddleware().process_request(req)



# -*- coding: utf-8 -*-


"""

Struct: worker_info
=================
>>> worker_info = {
        "id": int,
        "name": str,
        "arches": dict,
        "channels": dict,
        "enabled": bool,
        "max_load": bool,
        "ready": bool,
        "task_count": int,
        "current_load": float,
        "hub_version": str,
    }

Struct: task_info
=================
>>> task_info = {
        "id": int,
        "worker_id": int or None,
        "worker": worker_info or None,
        "parent": task_info or None,
        "state": str,
        "label": str,
        "method": str,
        "args": dict,
        "result": str,
        "exclusive": bool,
        "arch": dict,
        "channel": dict,
        "timeout": int,
        "waiting": bool,
        "awaited": bool,
        "dt_created": datetime,
        "dt_started": datetime or None,
        "dt_finished": datetime or None,
        "priority": int,
        "weight": int,
    }
"""

from __future__ import absolute_import

import datetime
import errno
import os
import signal
import sys
import time

from six import BytesIO
from six.moves.xmlrpc_client import Fault, ProtocolError

import six

import kobo.conf
import kobo.log
import kobo.tback
import kobo.worker.logger

from kobo.client import HubProxy
from kobo.client.constants import TASK_STATES
from kobo.exceptions import ShutdownException
from kobo.plugins import PluginContainer
from kobo.process import kill_process_group, get_process_status

from .task import FailTaskException


__all__ = (
    "TaskContainer",
    "TaskManager",
    "Fault",
    "ProtocolError",
    "ShutdownException",
)


# tasks are registered in __init__.py
class TaskContainer(PluginContainer):
    """Task container."""
    pass


class TaskManager(kobo.log.LoggingBase):
    """Task manager takes and executes new tasks."""

    def __init__(self, conf, logger=None, **kwargs):
        kobo.log.LoggingBase.__init__(self, logger)
        self.conf = kobo.conf.PyConfigParser()

        # load default config
        default_config = os.path.abspath(os.path.join(os.path.dirname(__file__), "default.conf"))
        self.conf.load_from_file(default_config)

        # update data from another config
        if conf is not None:
            self.conf.load_from_conf(conf)

        # update data from kwargs
        self.conf.load_from_dict(kwargs)

        self.pid_dict = {}  # { task_id: pid }
        self.task_dict = {}  # { task_id: { task information obtained from self.hub.get_worker_tasks() } }

        self.locked = False # if task manager is locked, it waits until tasks finish and exits

        self.task_container = TaskContainer()

        # self.hub (xml-rpc hub client) is created here
        self.hub = HubProxy(conf, client_type="worker", logger=self._logger, **kwargs)
        # worker information obtained from hub
        self.worker_info = self.hub.worker.get_worker_info()
        self.update_worker_info()

    def _task_str(self, task_info):
        """Return a task description."""
        return "#%s [%s]" % (task_info["id"], task_info["method"])

    def sleep(self):
        """Sleep between polls."""
        time.sleep(self.conf.get("SLEEP_TIME", 20))

    def update_worker_info(self):
        """Update worker_info dictionary."""
        self.log_debug("Updating worker info.")

        try:
            self.worker_info = self.hub.worker.update_worker(
                self.worker_info["enabled"],
                self.worker_info["ready"],
                len(self.pid_dict)
            )
        except ProtocolError as ex:
            self.log_error("Cannot update worker info: %s" % ex)
            return

    def wakeup_task(self, task_info):
        # alert is set in hub.worker.get_worker_tasks() when the task is supposed to wake up
        if task_info.get("alert", False):
            try:
                os.kill(self.pid_dict[task_info["id"]], signal.SIGUSR2)
            except OSError as ex:
                self.log_error("Cannot wake up task %s: %s" % (self._task_str(task_info), ex))
            else:
                self.log_info("Waking up task %s." % self._task_str(task_info))

    def update_tasks(self):
        """Read and process task statuses from hub.

        The processing we do is:
          1. clean up after tasks that are not longer active
          2. wake waiting tasks if appropriate
        """

        task_list = {}
        interrupted_list = []
        timeout_list = []
        finished_tasks = set()

        for task_info in self.hub.worker.get_worker_tasks():
            self.log_debug("Checking task: %s." % self._task_str(task_info))

            if task_info["state"] == TASK_STATES["OPEN"] and task_info["id"] not in self.pid_dict:
                # an interrupted task appears to be open, but running task manager doesn't track it
                # in it's pid list this happens after a power outage, for example
                interrupted_list.append(task_info["id"])
                finished_tasks.add(task_info["id"])
                continue

            if task_info["timeout"] is not None:
                time_delta = datetime.datetime.now() - datetime.datetime(*time.strptime(task_info["dt_started"], "%Y-%m-%d %H:%M:%S")[0:6])

                if time_delta.seconds >= (int(task_info["timeout"])):
                    timeout_list.append(task_info["id"])
                    finished_tasks.add(task_info["id"])
                    continue

            task_list[task_info["id"]] = task_info
            self.wakeup_task(task_info)

        self.task_dict = task_list
        self.log_debug("Current tasks: %r" % list(self.task_dict.keys()))

        if interrupted_list:
            self.log_warning("Closing interrupted tasks: %r" % sorted(interrupted_list))
            try:
                self.hub.worker.interrupt_tasks(interrupted_list)
            except (ShutdownException, KeyboardInterrupt):
                raise
            except Exception as ex:
                self.log_error("%s" % ex)

        if timeout_list:
            self.log_warning("Closing timed out tasks: %r" % sorted(timeout_list))
            try:
                self.hub.worker.timeout_tasks(timeout_list)
            except (ShutdownException, KeyboardInterrupt):
                raise
            except Exception as ex:
                self.log_error("%s" % ex)

        self.log_debug("pids: %s" % list(self.pid_dict.values()))

        for task_id in list(self.pid_dict.keys()):
            if self.is_finished_task(task_id):
                self.log_info("Task has finished: %s" % task_id)
                finished_tasks.add(task_id)
                # the subprocess handles most everything, we just need to clear things out
                if self.cleanup_task(task_id):
                    del self.pid_dict[task_id]
                if task_id in self.task_dict:
                    del self.task_dict[task_id]

        for task_id, pid in list(self.pid_dict.items()):
            if task_id not in self.task_dict:
                # expected to happen when:
                #  - we are in the narrow gap between the time the task records its result
                #    and the time the process actually exits.
                #  - task is canceled
                #  - task is forcibly reassigned/unassigned

                try:
                    task = self.hub.worker.get_task_no_verify(task_id)
                    if task["state"] == TASK_STATES["CANCELED"]:
                        self.log_info("Killing canceled task %r (pid %r)" % (task_id, pid))
                        if self.cleanup_task(task_id):
                            del self.pid_dict[task_id]
                            finished_tasks.add(task_id)
                    if task["state"] == TASK_STATES["TIMEOUT"]:
                        self.log_info("Killing timed out task %r (pid %r)" % (task_id, pid))
                        if self.cleanup_task(task_id):
                            del self.pid_dict[task_id]
                            finished_tasks.add(task_id)
                    elif "worker_id" in task and task["worker_id"] != self.worker_info["id"]:
                        self.log_info("Killing reassigned task %r (pid %r)" % (task_id, pid))
                        # TODO: Task.cleanup() - be careful, cleanup may remove running task's data!
                        if self.cleanup_task(task_id):
                            del self.pid_dict[task_id]
                    else:
                        self.log_warning("Lingering task %r (pid %r)" % (task_id, pid))
                except (ShutdownException, KeyboardInterrupt):
                    raise
                except Exception:
                    # TODO: do not catch generic error
                    self.log_error("Invalid task %r (pid %r)" % (task_id, pid))
                    raise

        for task_id in sorted(finished_tasks):
            task_info = self.hub.worker.get_task(task_id)
            self.finish_task(task_info)

        self.update_worker_info()

    def get_next_task(self):
        """Takes new task."""
        if not self.worker_info["enabled"]:
            self.log_info("Worker is disabled.")
            return

        if not self.worker_info["ready"]:
            self.log_info("Worker is not ready to take another task.")
            return

        if self.locked:
            task_list = self.hub.worker.get_worker_tasks()

            if not task_list:
                raise ShutdownException()

            awaited_task_list = self.hub.worker.get_awaited_tasks(task_list)
            self.log_debug("Current awaited tasks: %r" % [ti["id"] for ti in awaited_task_list])

            # process assigned tasks first
            for task_info in awaited_task_list:
                self.take_task(task_info)

            return

        assigned_task_list = self.hub.worker.get_tasks_to_assign()
        self.log_debug("Current assigned tasks: %r" % [ti["id"] for ti in assigned_task_list])

        # process assigned tasks first
        for task_info in assigned_task_list:
            self.take_task(task_info)

    def take_task(self, task_info):
        """Attempt to open the specified task. Return True on success, False otherwise."""

        if not self.worker_info["ready"]:
            return

        try:
            TaskClass = self.task_container[task_info["method"]]
        except (KeyError, ValueError):
            self.log_error("Cannot take unknown task %s (#%s)" % (task_info["method"], task_info["id"]))
            time.sleep(1) # prevent log flooding
            return

        if not TaskClass.exclusive:
            # always process exclusive tasks, regardless architecture or channel
            if task_info["arch"]["name"] not in TaskClass.arches:
                self.log_error("Unsupported arch for task %s: %s" % (self._task_str(task_info), task_info["arch"]["name"]))
                return

            if task_info["channel"]["name"] not in TaskClass.channels:
                self.log_error("Unsupported channel for task %s: %s)" % (self._task_str(task_info), task_info["channel"]["name"]))
                return

        self.log_info("Attempting to take task %s" % self._task_str(task_info))

        # TODO: improve exception handling and logging
        result = False
        reason = ""
        if task_info["state"] in (TASK_STATES["FREE"], TASK_STATES["ASSIGNED"]):
            try:
                # skip the "assign" step and try to open directly
                self.hub.worker.open_task(task_info["id"])
                result = True
            except (ShutdownException, KeyboardInterrupt):
                raise
            except Fault as ex:
                reason = "[%s] %s" % (ex.faultCode, ex.faultString)
            except Exception as ex:
                # TODO: log proper error message
                reason = "%s" % ex

        if not result:
            self.log_error("Cannot open task %s: %s" % (self._task_str(task_info), reason))
            return

        self.worker_info["current_load"] += TaskClass.weight
        self.worker_info["ready"] = self.worker_info["current_load"] < self.worker_info["max_load"]

        if TaskClass.foreground:
            self.run_task(task_info)
            self.finish_task(task_info)
        else:
            pid = self.fork_task(task_info)
            self.pid_dict[task_info["id"]] = pid

    def fork_task(self, task_info):
        self.log_debug("Forking task %s" % self._task_str(task_info))

        pid = os.fork()
        if pid:
            self.log_info("Task forked %s: pid=%s" % (self._task_str(task_info), pid))
            return pid

        # in no circumstance should we return after the fork
        # nor should any exceptions propagate past here
        try:
            # set process group
            os.setpgrp()

            # set a do-nothing handler for sigusr2
            # do not use signal.signal(signal.SIGUSR2, signal.SIG_IGN) - it completely masks interrups !!!
            signal.signal(signal.SIGUSR2, lambda *args: None)

            # set a default handler for SIGTERM
            signal.signal(signal.SIGTERM, signal.SIG_DFL)

            # run the task
            self.run_task(task_info)
        except Exception:
            self.log_critical("Error running forked task", exc_info=1)
            # don't bother raising since we're about to exit
        finally:
            # die
            os._exit(os.EX_OK)

    def run_task(self, task_info):
        TaskClass = self.task_container[task_info["method"]]

        # add *task_manager* attribute to foreground tasks
        if TaskClass.foreground:
            # TODO:
            TaskClass.task_manager = self
            hub = self.hub
        else:
            # create a new session for the task
            hub = HubProxy(self.conf, client_type="worker")

        task = TaskClass(hub, self.conf, task_info["id"], task_info["args"])

        # redirect stdout and stderr
        thread = kobo.worker.logger.LoggingThread(hub, task_info["id"])
        sys.stdout = kobo.worker.logger.LoggingIO(open(os.devnull, "w"), thread)
        sys.stderr = sys.stdout
        thread.start()

        failed = False
        try:
            task.run()
        except ShutdownException:
            thread.stop()
            if TaskClass.foreground and TaskClass.exclusive:
                # close task (shutdown-worker and similar control tasks) and raise
                hub.worker.close_task(task.task_id, task.result)
                raise
            # interrupt otherwise
            hub.worker.interrupt_tasks([task.task_id])
            return
        except KeyboardInterrupt:
            thread.stop()
            # interrupt otherwise
            hub.worker.interrupt_tasks([task.task_id])
            return
        except SystemExit as ex:
            if len(ex.args) > 0 and ex.args[0] != 0:
                sys.stdout.write("\nProgram has exited with return code '%s'." % ex.args[0])
                failed = True
        except FailTaskException as ex:
            failed = True
        except:
            message = "ERROR: %s\n" % kobo.tback.get_exception()
            message += "See traceback.log for details (admin only).\n"
            hub.upload_task_log(BytesIO(message.encode()), task.task_id, "error.log")
            hub.upload_task_log(BytesIO(kobo.tback.Traceback().get_traceback()), task.task_id, "traceback.log", mode=0o600)
            failed = True
        finally:
            thread.stop()

        if failed:
            hub.worker.fail_task(task.task_id, task.result)
        else:
            hub.worker.close_task(task.task_id, task.result)

    def finish_task(self, task_info):
        TaskClass = self.task_container[task_info["method"]]

        try:
            TaskClass.cleanup(self.hub, self.conf, task_info)
        except:
            self.log_critical(kobo.tback.get_exception())

        try:
            TaskClass.notification(self.hub, self.conf, task_info)
        except:
            self.log_critical(kobo.tback.get_exception())

    def is_finished_task(self, task_id):
        """Determine if task has finished.
        Calling os.waitpid removes finished child process zombies.
        """
        pid = self.pid_dict[task_id]

        try:
            (childpid, status) = os.waitpid(pid, os.WNOHANG)
        except OSError as ex:
            if ex.errno != errno.ECHILD:
                # should not happen
                self.log_error("Process hasn't exited with errno.ECHILD: %s" % task_id)
                raise

            # the process is already gone
            return False

        if childpid != 0:
            prefix = "Task #%s" % task_id
            self.log_info(get_process_status(status, prefix))
            return True

        return False

    def cleanup_task(self, task_id):
        """Cleanup after the task. Kill child processes."""

        try:
            success = kill_process_group(self.pid_dict[task_id], logger=self._logger)
        except IOError as ex:
            # proc file doesn"t exist -> process was already killed
            success = True

        if not success:
            try:
                success = kill_process_group(self.pid_dict[task_id], signal.SIGKILL, timeout=2, logger=self._logger)
            except IOError:
                # proc file doesn"t exist -> process was already killed
                success = True

        return success

    def shutdown(self):
        """Terminate all tasks and exit."""
        for task_id, task_info in six.iteritems(self.task_dict):
            try:
                TaskClass = self.task_container[task_info["method"]]
            except (AttributeError, ValueError):
                # unknown TaskClass -> skip it
                continue

            if not TaskClass.foreground:
                self.cleanup_task(task_id)

        if self.task_dict:
            # interrupt only if there are some tasks to interrupt
            self.hub.worker.interrupt_tasks(list(self.task_dict.keys()))
        self.update_worker_info()

    def lock(self):
        """Lock the task manager to finish all assigned tasks and exit."""
        self.locked = True
        self.log_info("Locking...")



# -*- coding: utf-8 -*-

import time

import unittest2 as unittest

from six import BytesIO

from mock import Mock

from kobo.worker.logger import LoggingThread, LoggingIO
from .utils import ArgumentIsInstanceOf


class TestLoggingThread(unittest.TestCase):

    def test_upload_task_log_on_stop(self):
        mock_hub = Mock()
        thread = LoggingThread(mock_hub, 9999)
        thread.daemon = True

        thread.start()
        self.assertTrue(thread.is_alive())
        self.assertTrue(thread._running)

        thread.write('This is a log message!')
        mock_hub.upload_task_log.assert_not_called()

        thread.stop()
        self.assertFalse(thread.is_alive())
        self.assertFalse(thread._running)
        mock_hub.upload_task_log.assert_called_once_with(ArgumentIsInstanceOf(BytesIO), 9999, 'stdout.log', append=True)

    def test_upload_task_log_after_some_time(self):
        mock_hub = Mock()
        thread = LoggingThread(mock_hub, 9999)
        thread.daemon = True

        thread.start()

        self.assertTrue(thread.is_alive())
        self.assertTrue(thread._running)

        for i in range(5):
            thread.write('%d - This is a very long message to be written in the log\n' % (i + 1))
            mock_hub.upload_task_log.assert_not_called()

        # let the thread running for a while
        time.sleep(.1)
        mock_hub.upload_task_log.assert_called_once_with(ArgumentIsInstanceOf(BytesIO), 9999, 'stdout.log', append=True)

        thread.stop()
        self.assertFalse(thread.is_alive())
        self.assertFalse(thread._running)


class TestLoggingIO(unittest.TestCase):

    def test_write(self):
        mock_io = Mock()
        mock_logging_thread = Mock()

        stream = LoggingIO(mock_io, mock_logging_thread)
        stream.write('This is a log message!')

        mock_io.write.assert_called_once_with('This is a log message!')
        mock_logging_thread.write.assert_called_once_with('This is a log message!')

    def test_write_multiple_calls(self):
        mock_io = Mock()
        mock_logging_thread = Mock()

        stream = LoggingIO(mock_io, mock_logging_thread)

        for i in range(10):
            stream.write('This is the %d log message!' % (i + 1))

        mock_io.write.assert_any_call('This is the 1 log message!')
        mock_io.write.assert_any_call('This is the 10 log message!')

        mock_logging_thread.write.assert_any_call('This is the 1 log message!')
        mock_logging_thread.write.assert_any_call('This is the 10 log message!')

        self.assertEqual(mock_io.write.call_count, 10)
        self.assertEqual(mock_logging_thread.write.call_count, 10)

    def test_getattr(self):
        mock_io = Mock()
        mock_logging_thread = Mock()

        stream = LoggingIO(mock_io, mock_logging_thread)
        random_variable = stream.random_variable

        self.assertFalse(random_variable is None)
        self.assertIsInstance(random_variable, Mock)



# -*- coding: utf-8 -*-

from __future__ import print_function
import sys
try:
    import json
except ImportError:
    import simplejson as json

from kobo.client import ClientCommand
from kobo.client.constants import TASK_STATES


class List_Tasks(ClientCommand):
    """list RUNNING and/or FREE tasks"""
    enabled = True


    def options(self):
        self.parser.usage = "%%prog %s [--free] [--running] [--verbose|--json]" % self.normalized_name

        self.parser.add_option(
            "--running",
            default=False,
            action="store_true",
            help="list RUNNING tasks"
        )

        self.parser.add_option(
            "--free",
            default=False,
            action="store_true",
            help="list FREE tasks",
        )

        self.parser.add_option(
            "--verbose",
            default=False,
            action="store_true",
            help="print details",
        )

        self.parser.add_option(
            "--json",
            default=False,
            action="store_true",
            help="print results in json",
        )

    def run(self, *args, **kwargs):
        username = kwargs.pop("username", None)
        password = kwargs.pop("password", None)
        verbose = kwargs.pop("verbose", False)
        use_json = kwargs.pop("json", False)

        if verbose and use_json:
            self.parser.error("It has no sense to use --verbose and --json in one time.")

        filters = []
        if kwargs['free']:
            filters += [TASK_STATES["FREE"], TASK_STATES["CREATED"]]
        if kwargs['running']:
            filters += [TASK_STATES["ASSIGNED"], TASK_STATES["OPEN"]]

        if not filters:
            self.parser.error("Use at least one from --free or --running options.")

        self.set_hub(username, password)
        result = sorted(self.hub.client.get_tasks([], filters), key=lambda x: x["id"])
        if use_json:
            print(json.dumps(result, indent=2, sort_keys=True))
        elif verbose:
            fmt = "%(id)8s %(state_label)-12s %(method)-20s %(owner)-12s %(worker)s"
            header = dict(id="TASKID", state_label="STATE", method="METHOD", owner="OWNER", worker="WORKER")
            print(fmt % header, file=sys.stderr)
            for task in result:
                print(fmt % task)
        else:
            for task in result:
                print(task['id'])



# -*- coding: utf-8 -*-


import os

from kobo.exceptions import ImproperlyConfigured
from django.conf import settings


for var in ["XMLRPC_METHODS", "TASK_DIR", "UPLOAD_DIR"]:
    if not hasattr(settings, var):
        raise ImproperlyConfigured("'%s' is missing in project settings. It must be set to run kobo.hub app." % var)


for var in ["TASK_DIR", "UPLOAD_DIR"]:
    dir_path = getattr(settings, var)
    if not os.path.isdir(dir_path):
        try:
            os.makedirs(dir_path)
        except:
            raise ImproperlyConfigured("'%s' doesn't exist and can't be automatically created." % dir_path)
    elif not os.access(dir_path, os.R_OK | os.W_OK | os.X_OK):
        raise ImproperlyConfigured("Invalid permissions on '%s'." % dir_path)


for var, value in [("MIDDLEWARE_CLASSES", "kobo.hub.middleware.WorkerMiddleware")]:
    if not hasattr(settings, var) or value not in getattr(settings, var, []):
        raise ImproperlyConfigured("'%s' in '%s' is missing in project settings. It must be set to run kobo.hub app." % (value, var))



# -*- coding: utf-8 -*-


try:
    import json
except ImportError:
    import simplejson as json

from kobo.client import ClientCommand


class Create_Task(ClientCommand):
    """create a new task which is either brand new or based on existing one"""
    enabled = True
    admin = True

    def options(self):
        self.parser.usage = "%%prog %s <options>" % self.normalized_name

        # if not specified, value from source task is used
        self.parser.add_option("--clone-from", dest="task_id", type="int", help="a task ID to clone a new task from")
        self.parser.add_option("--args", help="task arguments in JSON serialized dictionary")
        self.parser.add_option("--owner", dest="owner_name", help="username of the task owner")
        self.parser.add_option("--worker", dest="worker_name", help="name (hostname) of the worker")
        self.parser.add_option("--label", help="label or description")
        self.parser.add_option("--method", help="method name of the task handler")
        self.parser.add_option("--comment", help="comment")
        self.parser.add_option("--arch", dest="arch_name", help="arch name")
        self.parser.add_option("--channel", dest="channel_name", help="channel name")
        self.parser.add_option("--timeout", help="timeout")
        self.parser.add_option("--priority", help="priority")
        self.parser.add_option("--weight", help="weight")

    def _check_task_args(self, task_args):
        try:
            arg_dict = json.loads(task_args)
        except ValueError:
            self.parser.error("Arguments have to be specified in valid JSON.")

        if type(arg_dict) != dict:
            self.parser.error("Arguments have to be specified in dictionary type.")

    def run(self, *args, **kwargs):
        username = kwargs.pop("username", None)
        password = kwargs.pop("password", None)

        task_args = kwargs.get("args", None)
        if task_args != None:
            self._check_task_args(task_args)

        if kwargs["task_id"] is None:
            if kwargs["owner_name"] is None:
                self.parser.error("Owner is not set.")
            if kwargs["method"] is None:
                self.parser.error("Method is not set.")

        for key, value in kwargs.items():
            if value is None:
                del kwargs[key]

        self.set_hub(username, password)
        self.hub.client.create_task(kwargs)



# -*- coding: utf-8 -*-


"""
This module implements a simple threading worker pool.

Example:
    class MyThread(WorkerThread):
        def process(self, item, num):
            self.pool.log_debug("[%4s/%4s] Processing: %s" % (num, self.pool.queue_total, item))
            # process the *item*

    logger = ...
    pool = ThreadPool(logger=logger)

    # initialize threads
    for i in range(10):
        pool.add(MyThread(pool))

    pool.start()

    # populate the queue
    for i in range(100):
        pool.queue_put("item %s" % i)

    pool.stop()
"""


import sys
import threading
import six.moves.queue

import kobo.log
from six.moves import range
import six


class WorkerThread(threading.Thread):
    get_timeout = 1

    def __init__(self, pool, *args, **kwargs):
        threading.Thread.__init__(self, *args, **kwargs)
        self.pool = pool
        self.running = False
        self.kill = False
        self.failed = False

    def stop(self, kill=False):
        self.running = False # finish the work and join the thread
        self.kill = kill     # stop immediately without finishing the work
        self.join()

    def run(self):
        while (not self.kill) and (self.running or not self.pool.queue.empty()):
            try:
                item = self.pool.queue.get(timeout=self.get_timeout)
            except six.moves.queue.Empty:
                continue

            self.pool.queue_get_lock.acquire()
            self.pool.queue_processed += 1
            num = self.pool.queue_processed
            self.pool.queue_get_lock.release()

            try:
                self.process(item, num)
            except:
                self.failed = True
                self.pool.exceptions.append(sys.exc_info())
                self.pool.kill()

    def process(self, item, num):
        raise NotImplementedError


class ThreadPool(kobo.log.LoggingBase):
    def __init__(self, logger=None):
        kobo.log.LoggingBase.__init__(self, logger)
        self.threads = []
        self.exceptions = []
        self.queue = six.moves.queue.Queue()
        self.queue_put_lock = threading.Lock()
        self.queue_get_lock = threading.Lock()
        self.queue_total = 0
        self.queue_processed = 0

    def queue_put(self, item):
        """Put an item to the queue. Use this method instead of self.queue.put()."""
        self.queue.put(item)
        self.queue_put_lock.acquire()
        self.queue_total += 1
        self.queue_put_lock.release()

    def add(self, thread):
        self.threads.append(thread)

    def start(self):
        """Start all worker threads and immediately return."""
        for i in self.threads:
            i.running = True
            i.kill = False
            i.start()

    def stop(self):
        """Wait for the worker threads to process all items in queue.

        This method blocks until there is no more work left.

        It is essential to call this method. If you do any work in between
        ``start`` and ``stop``, make sure the ``stop`` method will always be
        called, otherwise the program will never exit.
        """
        for i in self.threads:
            i.running = False
        for i in self.threads:
            i.join()
        if self.exceptions:
            exc_info = self.exceptions[0]
            six.reraise(exc_info[0], exc_info[1], exc_info[2])

    def kill(self):
        for i in self.threads:
            i.kill = True
            i.running = False

def run_in_threads(func, params, threads=5, full_queue=True, use_lock=False):
    '''Run func with params (thread_task, param, queue_num) in threads
    - func: threadable function
    - params: iterable with arguments to func
    - threads: how many threads should be spawn
    - full_queue: start threads after filling queue - it will make value
                  of pool.queue_total same for all tasks. Otherwise pool will be started asap.
    - use_lock: add Lock objects for pool to func calls
    '''
    assert(threads > 0)
    if use_lock:
        lock = threading.Lock()
    class MyWorker(WorkerThread):
        def process(self, *args):
            if use_lock:
                args = list(args)
                args.append(lock)
            func(self, *args)
    pool = ThreadPool()
    if not full_queue:
        pool.start()
    for i in range(threads):
        pool.add(MyWorker(pool))
    for param in params:
        pool.queue_put(param)
    if full_queue:
        pool.start()
    pool.stop()



# -*- coding: utf-8 -*-


from __future__ import print_function
import sys

from kobo.client.task_watcher import TaskWatcher
from kobo.client import ClientCommand


class Resubmit_Tasks(ClientCommand):
    """resubmit failed tasks"""
    enabled = True


    def options(self):
        self.parser.usage = "%%prog %s task_id [task_id...]" % self.normalized_name
        self.parser.add_option("--force", action="store_true", help="Resubmit also tasks which are closed properly.")
        self.parser.add_option("--nowait", default=False, action="store_true", help="Don't wait until tasks finish.")


    def run(self, *args, **kwargs):
        if len(args) == 0:
            self.parser.error("At least one task id must be specified.")

        username = kwargs.pop("username", None)
        password = kwargs.pop("password", None)

        tasks = args

        self.set_hub(username, password)
        resubmitted_tasks = []
        failed = False
        for task_id in tasks:
            try:
                resubmitted_id = self.hub.client.resubmit_task(task_id, kwargs.pop("force", False))
                resubmitted_tasks.append(resubmitted_id)
            except Exception as ex:
                failed = True
                print(ex)

        if not kwargs.get('nowait'):
            TaskWatcher.watch_tasks(self.hub, resubmitted_tasks)
        if failed:
            sys.exit(1)



#!/usr/bin/python
# -*- coding: utf-8 -*-


import unittest2 as unittest

# tolerate and skip in absence of rpm since it's not installable to virtualenv
try:
    import rpm
except ImportError:
    HAVE_RPM = False
else:
    HAVE_RPM = True
    from kobo.rpmlib import parse_nvr, parse_nvra, compare_nvr, parse_evr, make_nvr, make_nvra, get_keys_from_header


@unittest.skipUnless(HAVE_RPM, "rpm python module is not installed")
class TestNVR(unittest.TestCase):
    def test_valid_nvr(self):
        self.assertEqual(parse_nvr("net-snmp-5.3.2.2-5.el5"), dict(name="net-snmp", version="5.3.2.2", release="5.el5", epoch=""))
        self.assertEqual(parse_nvr("1:net-snmp-5.3.2.2-5.el5"), dict(name="net-snmp", version="5.3.2.2", release="5.el5", epoch="1"))
        self.assertEqual(parse_nvr("net-snmp-1:5.3.2.2-5.el5"), dict(name="net-snmp", version="5.3.2.2", release="5.el5", epoch="1"))
        self.assertEqual(parse_nvr("net-snmp-5.3.2.2-5.el5:1"), dict(name="net-snmp", version="5.3.2.2", release="5.el5", epoch="1"))
        self.assertEqual(parse_nvr("/net-snmp-5.3.2.2-5.el5:1"), dict(name="net-snmp", version="5.3.2.2", release="5.el5", epoch="1"))
        self.assertEqual(parse_nvr("/1:net-snmp-5.3.2.2-5.el5"), dict(name="net-snmp", version="5.3.2.2", release="5.el5", epoch="1"))
        self.assertEqual(parse_nvr("foo/net-snmp-5.3.2.2-5.el5:1"), dict(name="net-snmp", version="5.3.2.2", release="5.el5", epoch="1"))
        self.assertEqual(parse_nvr("foo/1:net-snmp-5.3.2.2-5.el5"), dict(name="net-snmp", version="5.3.2.2", release="5.el5", epoch="1"))
        self.assertEqual(parse_nvr("/foo/bar/net-snmp-5.3.2.2-5.el5:1"), dict(name="net-snmp", version="5.3.2.2", release="5.el5", epoch="1"))
        self.assertEqual(parse_nvr("/foo/bar/1:net-snmp-5.3.2.2-5.el5"), dict(name="net-snmp", version="5.3.2.2", release="5.el5", epoch="1"))

        # test for name which contains the version number and a dash
        self.assertEqual(parse_nvr("openmpi-1.10-1.10.2-2.el6"), dict(name="openmpi-1.10", version="1.10.2", release="2.el6", epoch=""))

    def test_invalid_nvr(self):
        self.assertRaises(ValueError, parse_nvr, "net-snmp")
        self.assertRaises(ValueError, parse_nvr, "net-snmp-5.3.2.2-1:5.el5")
        self.assertRaises(ValueError, parse_nvr, "1:net-snmp-5.3.2.2-5.el5:1")
        self.assertRaises(ValueError, parse_nvr, "1:net-snmp-1:5.3.2.2-5.el5")
        self.assertRaises(ValueError, parse_nvr, "net-snmp-1:5.3.2.2-5.el5:1")
        self.assertRaises(ValueError, parse_nvr, "1:net-snmp-1:5.3.2.2-5.el5:1")

    def test_valid_nvra(self):
        self.assertEqual(parse_nvra("net-snmp-5.3.2.2-5.el5.i386"), dict(name="net-snmp", version="5.3.2.2", release="5.el5", epoch="", arch="i386", src=False))
        self.assertEqual(parse_nvra("net-snmp-5.3.2.2-5.el5.i386.rpm"), dict(name="net-snmp", version="5.3.2.2", release="5.el5", epoch="", arch="i386", src=False))
        self.assertEqual(parse_nvra("net-snmp-5.3.2.2-5.el5.src.rpm"), dict(name="net-snmp", version="5.3.2.2", release="5.el5", epoch="", arch="src", src=True))

        self.assertEqual(parse_nvra("/net-snmp-5.3.2.2-5.el5.src.rpm:1"), dict(name="net-snmp", version="5.3.2.2", release="5.el5", epoch="1", arch="src", src=True))
        self.assertEqual(parse_nvra("/1:net-snmp-5.3.2.2-5.el5.src.rpm"), dict(name="net-snmp", version="5.3.2.2", release="5.el5", epoch="1", arch="src", src=True))
        self.assertEqual(parse_nvra("foo/net-snmp-5.3.2.2-5.el5.src.rpm:1"), dict(name="net-snmp", version="5.3.2.2", release="5.el5", epoch="1", arch="src", src=True))
        self.assertEqual(parse_nvra("foo/1:net-snmp-5.3.2.2-5.el5.src.rpm"), dict(name="net-snmp", version="5.3.2.2", release="5.el5", epoch="1", arch="src", src=True))
        self.assertEqual(parse_nvra("/foo/bar/net-snmp-5.3.2.2-5.el5.src.rpm:1"), dict(name="net-snmp", version="5.3.2.2", release="5.el5", epoch="1", arch="src", src=True))
        self.assertEqual(parse_nvra("/foo/bar/1:net-snmp-5.3.2.2-5.el5.src.rpm"), dict(name="net-snmp", version="5.3.2.2", release="5.el5", epoch="1", arch="src", src=True))

    def test_invalid_nvra(self):
        self.assertEqual(parse_nvra("net-snmp-5.3.2.2-5.el5.i386.rpm:1"), dict(name="net-snmp", version="5.3.2.2", release="5.el5", epoch="1", arch="i386", src=False))
        self.assertEqual(parse_nvra("net-snmp-5.3.2.2-5.el5.i386:1.rpm"), dict(name="net-snmp", version="5.3.2.2", release="5.el5", epoch="1", arch="i386", src=False))
        self.assertRaises(ValueError, parse_nvra, "net-snmp-5.3.2.2-5")

    def test_compare_nvr(self):
        first = {'name': 'a', 'version': '1', 'release': '1', 'epoch': '1'}
        second = {'name': 'a', 'version': '1', 'release': '1', 'epoch': '1'}
        self.assertEqual(compare_nvr(first, second), 0)
        second['version'] = '0'
        self.assertEqual(compare_nvr(first, second), 1)
        second['version'] = 0
        self.assertEqual(compare_nvr(first, second), 1)
        second['version'] = 2
        self.assertEqual(compare_nvr(first, second), -1)
        second['version'] = 1
        second['release'] = 0
        self.assertEqual(compare_nvr(first, second), 1)
        second['release'] = 2
        self.assertEqual(compare_nvr(first, second), -1)
        second['release'] = 1
        second['epoch'] = 0
        self.assertEqual(compare_nvr(first, second), 1)
        second['epoch'] = 2
        self.assertEqual(compare_nvr(first, second), -1)
        first = {'name': 'a', 'version': '1', 'release': '1', 'epoch': None}
        second = {'name': 'a', 'version': '1', 'release': '1', 'epoch': '1'}
        self.assertEqual(compare_nvr(first, second), -1)
        # missing epoch
        first = {'name': 'a', 'version': '1', 'release': '1'}
        self.assertEqual(compare_nvr(first, second), -1)

    def test_valid_evr(self):
        self.assertEqual(parse_evr("5.3.2.2-5.el5"), {"epoch": "", "version": "5.3.2.2", "release": "5.el5"})
        self.assertEqual(parse_evr("1:5.3.2.2-5.el5"), {"epoch": "1", "version": "5.3.2.2", "release": "5.el5"})
        self.assertEqual(parse_evr("5.3.2.2-5.el5:1"), {"epoch": "1", "version": "5.3.2.2", "release": "5.el5"})
        self.assertEqual(parse_evr("5.3.2.2:1", allow_empty_release=True), {"epoch": "1", "version": "5.3.2.2", "release": ""})
        self.assertEqual(parse_evr("1:5.3.2.2", allow_empty_release=True), {"epoch": "1", "version": "5.3.2.2", "release": ""})
        self.assertEqual(parse_evr("1:5", allow_empty_release=True), {"epoch": "1", "version": "5", "release": ""})
        self.assertEqual(parse_evr("1", allow_empty_release=True), {"epoch": "", "version": "1", "release": ""})

    def test_invalid_evr(self):
        self.assertRaises(ValueError, parse_evr, "a:b")
        self.assertRaises(ValueError, parse_evr, "5.3.2.2:1", allow_empty_release=False)
        self.assertRaises(ValueError, parse_evr, "1:5.3.2.2", allow_empty_release=False)
        self.assertRaises(ValueError, parse_evr, "1:5", allow_empty_release=False)
        self.assertRaises(ValueError, parse_evr, "1", allow_empty_release=False)

    def test_make_nvr(self):
        nvr = dict(name="net-snmp", version="5.3.2.2", release="5.el5")
        self.assertEqual(make_nvr(nvr), "net-snmp-5.3.2.2-5.el5")

        nvr = dict(name="net-snmp", version="5.3.2.2", release="5.el5")
        self.assertEqual(make_nvr(nvr), "net-snmp-5.3.2.2-5.el5")
        self.assertEqual(make_nvr(nvr, add_epoch=True), "net-snmp-5.3.2.2-5.el5")
        self.assertEqual(make_nvr(nvr, force_epoch=True), "net-snmp-0:5.3.2.2-5.el5")
        self.assertEqual(make_nvr(nvr, add_epoch=True, force_epoch=True), "net-snmp-0:5.3.2.2-5.el5")
        # force_epoch overrides add_epoch
        self.assertEqual(make_nvr(nvr, add_epoch=False, force_epoch=True), "net-snmp-0:5.3.2.2-5.el5")

        nvr["epoch"] = None
        self.assertEqual(make_nvr(nvr, add_epoch=True), "net-snmp-5.3.2.2-5.el5")
        self.assertEqual(make_nvr(nvr, add_epoch=True, force_epoch=True), "net-snmp-0:5.3.2.2-5.el5")

        nvr["epoch"] = ""
        self.assertEqual(make_nvr(nvr, add_epoch=True), "net-snmp-5.3.2.2-5.el5")
        self.assertEqual(make_nvr(nvr, add_epoch=True, force_epoch=True), "net-snmp-0:5.3.2.2-5.el5")

        nvr["epoch"] = "0"
        self.assertEqual(make_nvr(nvr, add_epoch=True), "net-snmp-5.3.2.2-5.el5")
        self.assertEqual(make_nvr(nvr, add_epoch=True, force_epoch=True), "net-snmp-0:5.3.2.2-5.el5")

        nvr["epoch"] = 0
        self.assertEqual(make_nvr(nvr, add_epoch=True), "net-snmp-5.3.2.2-5.el5")
        self.assertEqual(make_nvr(nvr, add_epoch=True, force_epoch=True), "net-snmp-0:5.3.2.2-5.el5")

        nvr["epoch"] = 1
        self.assertEqual(make_nvr(nvr, add_epoch=True), "net-snmp-1:5.3.2.2-5.el5")
        self.assertEqual(make_nvr(nvr, add_epoch=True, force_epoch=True), "net-snmp-1:5.3.2.2-5.el5")

    def test_make_nvra(self):
        nvra = dict(name="net-snmp", version="5.3.2.2", release="5.el5", arch="i386")
        self.assertEqual(make_nvra(nvra), "net-snmp-5.3.2.2-5.el5.i386")
        self.assertEqual(make_nvra(nvra, add_rpm=True), "net-snmp-5.3.2.2-5.el5.i386.rpm")
        self.assertEqual(make_nvra(nvra, add_epoch=True), "net-snmp-5.3.2.2-5.el5.i386")
        self.assertEqual(make_nvra(nvra, force_epoch=True), "net-snmp-0:5.3.2.2-5.el5.i386")
        self.assertEqual(make_nvra(nvra, add_epoch=True, force_epoch=True), "net-snmp-0:5.3.2.2-5.el5.i386")
        # force_epoch overrides add_epoch
        self.assertEqual(make_nvra(nvra, add_epoch=False, force_epoch=True), "net-snmp-0:5.3.2.2-5.el5.i386")
        self.assertEqual(make_nvra(nvra, add_epoch=True, force_epoch=True, add_rpm=True), "net-snmp-0:5.3.2.2-5.el5.i386.rpm")

        nvra["epoch"] = None
        self.assertEqual(make_nvra(nvra, add_epoch=True), "net-snmp-5.3.2.2-5.el5.i386")
        self.assertEqual(make_nvra(nvra, add_epoch=True, force_epoch=True), "net-snmp-0:5.3.2.2-5.el5.i386")

        nvra["epoch"] = ""
        self.assertEqual(make_nvra(nvra, add_epoch=True), "net-snmp-5.3.2.2-5.el5.i386")
        self.assertEqual(make_nvra(nvra, add_epoch=True, force_epoch=True), "net-snmp-0:5.3.2.2-5.el5.i386")

        nvra["epoch"] = "0"
        self.assertEqual(make_nvra(nvra, add_epoch=True), "net-snmp-5.3.2.2-5.el5.i386")
        self.assertEqual(make_nvra(nvra, add_epoch=True, force_epoch=True), "net-snmp-0:5.3.2.2-5.el5.i386")

        nvra["epoch"] = 0
        self.assertEqual(make_nvra(nvra, add_epoch=True), "net-snmp-5.3.2.2-5.el5.i386")
        self.assertEqual(make_nvra(nvra, add_epoch=True, force_epoch=True), "net-snmp-0:5.3.2.2-5.el5.i386")

        nvra["epoch"] = 1
        self.assertEqual(make_nvra(nvra, add_epoch=True), "net-snmp-1:5.3.2.2-5.el5.i386")
        self.assertEqual(make_nvra(nvra, add_epoch=True, force_epoch=True), "net-snmp-1:5.3.2.2-5.el5.i386")

        nvra = dict(name="openmpi-1.10", version="1.10.2", release="2.el6", arch="i386")
        self.assertEqual(make_nvra(nvra), "openmpi-1.10-1.10.2-2.el6.i386")


@unittest.skipUnless(HAVE_RPM, "rpm python module is not installed")
class TestGetKeys(unittest.TestCase):
    def test_sigkey(self):
        header = {
            rpm.RPMTAG_DSAHEADER: None,
            rpm.RPMTAG_RSAHEADER: b"\x89\x02\x15\x03\x05\x00]\xb2\x95\\\xef<\x11\x1f\xcf\xc6Y\xb9\x01\x08E.\x0f\xffC\x10\xe5\x0e\xff\xa3\x8a\xce\xfb\xff\xf1\xf6\x98L|\xf5\xcbd\x8dc\xa8IsP\x18\x85\xc0\x06\xc0L\xff\x95\xd7\x86\xdd\xbe\xf1\xe0L%P2^\x14\xfe1\xf6\x85\x88\xe7\x93\xd6{\xaa\xf0\x1d\xa7YJ\x8e\xfc\x08\xa0\x9f\x8e\x10|\xd7(\x94\n\xdd\x06Ta\xca\x94\x0e\x05\x1f\xb2jv|JB^\xffAU\x90\x96d\x88K\xcb<\x06\xca\x93\x7f\xc2B@\xa2\x10DF\x8e\x04\xa5\xf9C\x86\xe3Z.\n\xe4\xde\xa6\x83\xa1\x85\x91I#o\xa4\x8e{\xbc\xdd\xcdF\xa35\xc2~\xed.\xfb\x18\xecS\x95\xb7iz\xfd=\x1d\x9a\xd9h\x7fP@\xcb\xfbW-\x12\xfc\xa2f\xf9\x7fd\xe1\x87\xedu\x9bTZ\\\xbc9\xea\xad\xee\xf8\xdf\xb7f\x88\xdc$\xad\x1c/\xd7\xc1\x1b\xca\xc0\x1b\xd7q\xaa\x13\xa4J\x92\x971X\xac\x16I\xd1\xde\x011+jW\x08\x9dC\x86\x00\x02!g%@h\xe6u\xe1\xd5e\xa7j\x80\xe2\x9bH\xdb\xec\x84\xca \x80\x8c\xbaQ4\xfdVp1\xe1\x98\xe2\xe7\xfb\xdf\x0cE6\xe2\xa7\n^\x14\xe4\xe1\xb2\x06\x1fU\xe7:\xcdH\x15\xa8\x10[\xdd\xf5\xa2B\x1f\xdf\xf9\x9e\x8au\x93l\x97\xb7\x12\xe9\xf9'\x0e\xd4\xe2a\xd6\xa4\xbc4\xa4y\x88\xe8\xabE\xe3\xbf\x15uaFs\x08\xad\xd5,c4K\xa8%ZMHmy\\_\x05\xc6\x0f\xdc0\xcc\x80B\xda\xde\xf0A\x1b\xeb\xe4\x13\x9e\xad\x9d)a\xd4I\xd7\xfd\xc7,\xc5\xf2{\xcc\x0b\t&\xce2\xfb\xe89\x91< \xe1a\xbd\xec\xe6\xe0-K\xc2DX\xc3\xbf\x95\xf3\xf1\x9c\xaa\x91f\xb2}\x9bj+\xb7\x86:\x99\x81\xfd\x96\xbb3\x02\x90\x05\xf6\xa8\x1e#<Na\xdf*\xab\xc9\xd8\x1a\xeaA7bR\xec4/\x83\xf0\x13\x1a\x80%\xe0#\x89\xd0\xa5\x0f\xb8\xb7\x8bNG\x9c~\x9c\xae\xe2\xc5\xeb\x17K\xed\x8d@\x12{\xb0\xe4\xfc\xb1\xe4\xda\xce\xe2\xc7\xe6\xdf@\xd0\x05\xde\x153b\x08\x1b\x80\x11\x80Z\x95A\x85+sQ _\xcd\xee/\xa8\x0fx\xd8\xad\r}\xb4\x19",
            rpm.RPMTAG_SIGGPG: None,
            rpm.RPMTAG_SIGPGP: b'\x89\x02\x15\x03\x05\x00]\xb2\x95\\\xef<\x11\x1f\xcf\xc6Y\xb9\x01\x08+P\x0f\xfd\x1b\x06\x11\xb5\xf2\xcf\x97\xa7+\xd2\x0biU_\xda\xd1\xb4\xf4\xd6A\xf55\xc7se\x92<\xffC\xdc,\xfa\xa6&xbm\xfe\x943f\xedbe \xac\xf9\x1d\x91\xd4\x97\xdd~v\xb5\xf9\x14\xc2\xdc\xdb\xb2\xe0\x08\\g\xaeS&\xed\x1f$U{\xc4\x1f\xe8\xc3\xdfF\xb3\xa0\xdb\xcbo\x94_L\xb0\xf1\x8d<\xa1c\xe3\x16f7E\xd3\xa6O\xde\xa3NG\x10\xbc\x0b*\xaa\xd2\xe2\xe2\xf9\xc2\xb8\xb6\xee\xa5\x88s\xc5\xb7\xcc=\x9e\x17\x97~\x1c\xa4x\x06ri\x80-\x08@ l0\xce\xe2k\xb7\xe3h}\x86\x0c\x88q\x80(\xa6\xc6\xa1\xf1Q\xfd0\xbbg\xf8\xe4\xe1\xe2t\x00O\'\x07D\xdfZ\x10\xd6\x1d\xf0\xee\x1e\xc7\x074\x0e\xb5\xf9\xacO\x89\x90M`\xb1F\xfa\\b\x83\x96\xb0\xcf\xc4\x9b\xdd\xf8$\xc4\x92\x83\x87\x15(6)H\xa0\xdf!\xdc\x04\xdb\x15\x81S%\x0bf\xba\xa6T\x1c\x02_\x08\xfe\xd6]Aa\x81\x8f\xe2Y"\xe0\x08\xb3\xa3\xfak*V\xa3\xdc\xab\xbd\xb3=T\x8fu\x7fN9\xf2s&\xde\x0512\x9b\xe4\xdcY\xeeP\xf1\xd8\xe8\x1dB\xa5v\xb3\xc6\xd8k\x92[\xae\xa2\x8e*\xd9r\x02P2\xf9Vpg\x8b\xd4\xac\xca&\xb2\x13\xc1\x8fu\xee\x96\xce\x02UZ\xc4-O\xae\x04f\x03\xab$u\xb4\x17\xcf[\x00\xc3U\xe4\xe2\xf1$\xcb\xfe\x8a\xa3\x08O\x06\x7fVf%+Dg\x1b\xaa\x06!\xdc\xb5\xb5\xb6\xec\xf3$\xc9\xd0\x1c\x89\x0b\xd4IN\x03\x07\xcd^\x97\\\xce\x93\xce\xeb\xf1\xae\xacB\xe0\x04A\x0b\xbe!\x8d\xa7.W\xa4\xfa\xea[\xebj7\t\xa8\xda\xd0\xff(\x00a\xb0V\x04\x8f\x93 \x1eiQ\x9dG\xb7\x04g\xbc\x86\xc2\x0f\x06\xf4\x83N\xfc\x9e\x88\xb3D\xff\x92\xd9E\xd7\xc6\xea\\Z\x08^\xf3\xd2\xbf\x83\xfe\xef_\xe5|?\xc5x\xc1\xfd\xbd\x8c&,\x17\x1a\x1c\xf1<c\x0b6C\x8b\x7f\xa0|\x9a\xa37\xf9\xf8U\x90\x1db\x85=\xd4\x0e1\xe4\xdf\xba$\xe3\x05=B!4D\xc8\xc9\x15-',
        }
        self.assertEqual(get_keys_from_header(header), "CFC659B9")



# -*- coding: utf-8 -*-

import errno
import os
import signal
import logging

import django
import pytest
import six

# Only for Django >= 1.7
if 'setup' in dir(django):
    # This has to happen before below imports because they have a hard requirement
    # on settings being loaded before import.
    django.setup()

from datetime import datetime, timedelta

from django.contrib.auth.models import User

from mock import Mock, PropertyMock, patch

from kobo.client.constants import TASK_STATES
from kobo.exceptions import ShutdownException
from kobo.hub.models import Arch, Channel, Task, Worker
from kobo.worker import TaskBase
from kobo.worker.task import FailTaskException
from kobo.worker.taskmanager import TaskManager, TaskContainer
from six.moves.xmlrpc_client import ProtocolError

from .rpc import HubProxyMock, RpcServiceMock
from .utils import DjangoRunner

runner = DjangoRunner()
setup_module = runner.start
teardown_module = runner.stop


class DummyTask(TaskBase):

    def run(self):
        self.result = True

    @classmethod
    def cleanup(cls, hub, conf, task_info):
        pass

    @classmethod
    def notification(cls, hub, conf, task_info):
        pass


class DummyForegroundTask(DummyTask):
    enabled = True
    arches = ['testarch']
    channels = ['testchannel']
    exclusive = False
    foreground = True
    priority = 10
    weight = 1.0


class DummyForkTask(DummyTask):
    enabled = True
    arches = ['testarch']
    channels = ['testchannel']
    exclusive = False
    foreground = False
    priority = 10
    weight = 1.0


class DummyHeavyTask(DummyTask):
    enabled = True
    arches = ['testarch']
    channels = ['testchannel']
    exclusive = False
    foreground = False
    priority = 10
    weight = 100.0


class DummyFailTask(DummyTask):
    enabled = True
    arches = ['testarch']
    channels = ['testchannel']
    exclusive = False
    foreground = True
    priority = 10
    weight = 1.0

    def run(self):
        raise FailTaskException()


class DummyExceptionTask(DummyTask):
    enabled = True
    arches = ['testarch']
    channels = ['testchannel']
    exclusive = False
    foreground = True
    priority = 10
    weight = 1.0

    def run(self):
        raise Exception()


class TestTaskManager(django.test.TransactionTestCase):

    def setUp(self):
        self._fixture_teardown()
        super(TestTaskManager, self).setUp()

        TaskContainer.register_plugin(DummyForegroundTask)
        TaskContainer.register_plugin(DummyForkTask)
        TaskContainer.register_plugin(DummyHeavyTask)
        TaskContainer.register_plugin(DummyFailTask)
        TaskContainer.register_plugin(DummyExceptionTask)

        user = User.objects.create(username='testuser')
        arch = Arch.objects.create(name='testarch')
        channel = Channel.objects.create(name='testchannel')

        w = Worker.objects.create(
            worker_key='mock-worker',
            name='mock-worker',
        )

        w.arches.add(arch)
        w.channels.add(channel)
        w.save()

        self._arch = arch
        self._channel = channel
        self._user = user
        self._worker = RpcServiceMock(w)

    @pytest.mark.xfail(reason='Check issue #68 for more info (https://git.io/fxSZ2).')
    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_update_worker_info(self):
        tm = TaskManager(conf={'worker': self._worker})
        self.assertTrue(tm.worker_info['enabled'])
        self.assertTrue(tm.worker_info['ready'])

        tm.worker_info['enabled'] = False
        tm.worker_info['ready'] = False
        tm.update_worker_info()
        self.assertFalse(tm.worker_info['enabled'])
        self.assertFalse(tm.worker_info['ready'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_update_worker_info_catch_protocol_error(self):
        tm = TaskManager(conf={'worker': self._worker})

        tm.hub.worker.update_worker = Mock(side_effect=ProtocolError(None, None, None, None))
        tm.update_worker_info()

        with self.assertRaises(ValueError):
            tm.hub.worker.update_worker = Mock(side_effect=ValueError)
            tm.update_worker_info()

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_wakeup_task_if_alert_is_set(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForkTask',
            state=TASK_STATES['FREE'],
        )

        tm = TaskManager(conf={'worker': self._worker})

        task_info = t.export(False)
        task_info['alert'] = True

        with patch('kobo.worker.taskmanager.os', fork=Mock(return_value=9999)) as os_mock:
            tm.take_task(task_info)
            os_mock.fork.assert_called_once()
            tm.wakeup_task(task_info)
            os_mock.kill.assert_called_once_with(9999, signal.SIGUSR2)

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_wakeup_task_if_alert_is_not_set(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForkTask',
            state=TASK_STATES['FREE'],
        )

        tm = TaskManager(conf={'worker': self._worker})
        task_info = t.export(False)

        with patch('kobo.worker.taskmanager.os', fork=Mock(return_value=9999)) as os_mock:
            tm.take_task(task_info)
            os_mock.fork.assert_called_once()
            tm.wakeup_task(task_info)
            os_mock.kill.assert_not_called()

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_wakeup_task_catch_os_error(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForkTask',
            state=TASK_STATES['FREE'],
        )

        tm = TaskManager(conf={'worker': self._worker})

        task_info = t.export(False)
        task_info['alert'] = True

        with patch('kobo.worker.taskmanager.os', fork=Mock(return_value=9999)) as os_mock:
            tm.take_task(task_info)

        with patch('kobo.worker.taskmanager.os', kill=Mock(side_effect=OSError)) as os_mock:
            tm.wakeup_task(task_info)
            os_mock.kill.assert_called_once_with(9999, signal.SIGUSR2)

        with self.assertRaises(ValueError):
            with patch('kobo.worker.taskmanager.os', kill=Mock(side_effect=ValueError)) as os_mock:
                tm.wakeup_task(task_info)
                os_mock.kill.assert_called_once_with(9999, signal.SIGUSR2)

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_update_tasks_timeout_task_if_running_for_to_long(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForkTask',
            timeout=-1, # task will always timeout with negative value
            state=TASK_STATES['FREE'],
        )

        self.assertEqual(t.state, TASK_STATES['FREE'])

        tm = TaskManager(conf={'worker': self._worker})
        task_info = t.export(False)

        with patch('kobo.worker.taskmanager.os', fork=Mock(return_value=9999)) as os_mock:
            tm.take_task(task_info)
            os_mock.fork.assert_called_once()

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['OPEN'])

        self.assertTrue(t.id in tm.pid_dict)
        tm.update_tasks()
        self.assertFalse(t.id in tm.pid_dict)

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['TIMEOUT'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_update_tasks_finish_task_if_canceled(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForkTask',
            state=TASK_STATES['FREE'],
        )

        self.assertEqual(t.state, TASK_STATES['FREE'])

        tm = TaskManager(conf={'worker': self._worker})
        task_info = t.export(False)

        with patch('kobo.worker.taskmanager.os', fork=Mock(return_value=9999)) as os_mock:
            tm.take_task(task_info)
            os_mock.fork.assert_called_once()

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['OPEN'])

        self._worker.cancel_task(t.id)

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['CANCELED'])

        self.assertTrue(t.id in tm.pid_dict)
        tm.update_tasks()
        self.assertFalse(t.id in tm.pid_dict)

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['CANCELED'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_update_tasks_interrupt_taks_if_not_on_pid_list(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForegroundTask',
            state=TASK_STATES['OPEN'],
        )

        self.assertEqual(t.state, TASK_STATES['OPEN'])

        tm = TaskManager(conf={'worker': self._worker})

        self.assertFalse(t.id in tm.pid_dict)
        tm.update_tasks()
        self.assertFalse(t.id in tm.pid_dict)

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['INTERRUPTED'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_update_tasks_cleanup_finished_tasks(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForkTask',
            state=TASK_STATES['FREE'],
        )

        self.assertEqual(t.state, TASK_STATES['FREE'])

        tm = TaskManager(conf={'worker': self._worker})
        task_info = t.export(False)

        with patch('kobo.worker.taskmanager.os', fork=Mock(return_value=9999)) as os_mock:
            tm.take_task(task_info)
            os_mock.fork.assert_called_once()

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['OPEN'])

        tm.run_task(task_info)

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['CLOSED'])

        with patch('kobo.worker.taskmanager.os', waitpid=Mock(return_value=(123, 0))) as os_mock:
            self.assertTrue(t.id in tm.pid_dict)
            tm.update_tasks()
            self.assertFalse(t.id in tm.pid_dict)
            os_mock.waitpid.assert_called_once()

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['CLOSED'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_update_tasks_without_tasks(self):
        tm = TaskManager(conf={'worker': self._worker})
        tm.update_tasks()
        self.assertEqual(len(tm.pid_dict.keys()), 0)
        self.assertEqual(len(tm.task_dict.keys()), 0)

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_get_next_task_runs_free_task(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForegroundTask',
            state=TASK_STATES['FREE'],
        )

        self.assertEqual(t.state, TASK_STATES['FREE'])

        tm = TaskManager(conf={'worker': self._worker})
        tm.get_next_task()

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['CLOSED'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_get_next_task_dont_run_tasks_if_disabled(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForegroundTask',
            state=TASK_STATES['FREE'],
        )

        self.assertEqual(t.state, TASK_STATES['FREE'])

        tm = TaskManager(conf={'worker': self._worker})
        tm.worker_info['enabled'] = False

        tm.get_next_task()

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['FREE'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_get_next_task_dont_run_tasks_if_not_ready(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForegroundTask',
            state=TASK_STATES['FREE'],
        )

        self.assertEqual(t.state, TASK_STATES['FREE'])

        tm = TaskManager(conf={'worker': self._worker})
        tm.worker_info['ready'] = False

        tm.get_next_task()

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['FREE'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_get_next_task_raise_shutdown_exception_if_locked_and_no_tasks_running(self):
        tm = TaskManager(conf={'worker': self._worker})
        tm.lock()

        # ensure no tasks are running
        self.assertEqual(len(self._worker.worker.running_tasks()), 0)

        with self.assertRaises(ShutdownException):
            tm.get_next_task()

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_get_next_task_runs_assigened_awaited_task_if_locked(self):
        t1 = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForegroundTask',
            state=TASK_STATES['OPEN'],
        )

        t2 = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForegroundTask',
            state=TASK_STATES['ASSIGNED'],
            awaited=True,
            parent=t1,
        )

        t3 = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForegroundTask',
            waiting=True,
            parent=t2,
            state=TASK_STATES['FREE'],
        )

        self.assertEqual(t1.state, TASK_STATES['OPEN'])
        self.assertEqual(t2.state, TASK_STATES['ASSIGNED'])
        self.assertEqual(t3.state, TASK_STATES['FREE'])

        tm = TaskManager(conf={'worker': self._worker})
        tm.lock()
        tm.get_next_task()

        # reload task info
        t1 = Task.objects.get(id=t1.id)
        self.assertEqual(t1.state, TASK_STATES['OPEN'])

        t2 = Task.objects.get(id=t2.id)
        self.assertEqual(t2.state, TASK_STATES['CLOSED'])

        t3 = Task.objects.get(id=t3.id)
        self.assertEqual(t3.state, TASK_STATES['FREE'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_take_task_open_task_if_free(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForkTask',
            state=TASK_STATES['FREE'],
        )

        self.assertEqual(t.state, TASK_STATES['FREE'])

        tm = TaskManager(conf={'worker': self._worker})
        task_info = t.export(False)

        with patch('kobo.worker.taskmanager.os', fork=Mock(return_value=9999)) as os_mock:
            tm.take_task(task_info)
            os_mock.fork.assert_called_once()

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['OPEN'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_take_task_open_task_if_assigned(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForkTask',
            state=TASK_STATES['ASSIGNED'],
        )

        self.assertEqual(t.state, TASK_STATES['ASSIGNED'])

        tm = TaskManager(conf={'worker': self._worker})
        task_info = t.export(False)

        with patch('kobo.worker.taskmanager.os', fork=Mock(return_value=9999)) as os_mock:
            tm.take_task(task_info)
            os_mock.fork.assert_called_once()

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['OPEN'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_take_task_runs_task_if_foreground(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForegroundTask',
            state=TASK_STATES['FREE'],
        )

        self.assertEqual(t.state, TASK_STATES['FREE'])

        tm = TaskManager(conf={'worker': self._worker})
        task_info = t.export(False)

        with patch('kobo.worker.taskmanager.os', fork=Mock(return_value=9999)) as os_mock:
            os_mock.devnull = os.devnull

            tm.take_task(task_info)
            os_mock.fork.assert_not_called()

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['CLOSED'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_take_task_catch_errors_if_cant_open_task(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForegroundTask',
            state=TASK_STATES['FREE'],
        )

        self.assertEqual(t.state, TASK_STATES['FREE'])

        tm = TaskManager(conf={'worker': self._worker})
        tm.hub.worker.open_task = Mock(side_effect=ValueError)

        task_info = t.export(False)

        with patch('kobo.worker.taskmanager.os', fork=Mock(return_value=9999)) as os_mock:
            tm.take_task(task_info)
            os_mock.fork.assert_not_called()

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['FREE'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_take_task_if_worker_not_ready(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForkTask',
            state=TASK_STATES['FREE'],
        )

        self.assertEqual(t.state, TASK_STATES['FREE'])

        tm = TaskManager(conf={'worker': self._worker})
        tm.worker_info['ready'] = False

        task_info = t.export(False)

        with patch('kobo.worker.taskmanager.os', fork=Mock(return_value=9999)) as os_mock:
            tm.take_task(task_info)
            os_mock.fork.assert_not_called()

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['FREE'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_take_task_if_wrong_arch(self):
        arch = Arch.objects.create(name='arch_x86', pretty_name='Arch x86')

        t = Task.objects.create(
            worker=self._worker.worker,
            arch=arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForkTask',
            state=TASK_STATES['FREE'],
        )

        self.assertEqual(t.state, TASK_STATES['FREE'])

        tm = TaskManager(conf={'worker': self._worker})
        task_info = t.export(False)

        with patch('kobo.worker.taskmanager.os', fork=Mock(return_value=9999)) as os_mock:
            tm.take_task(task_info)
            os_mock.fork.assert_not_called()

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['FREE'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_take_task_if_wrong_channel(self):
        channel = Channel.objects.create(name='channel_x86')

        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=channel,
            owner=self._user,
            method='DummyForkTask',
            state=TASK_STATES['FREE'],
        )

        self.assertEqual(t.state, TASK_STATES['FREE'])

        tm = TaskManager(conf={'worker': self._worker})
        task_info = t.export(False)

        with patch('kobo.worker.taskmanager.os', fork=Mock(return_value=9999)) as os_mock:
            tm.take_task(task_info)
            os_mock.fork.assert_not_called()

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['FREE'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_take_task_set_load_and_ready(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyHeavyTask',
            state=TASK_STATES['FREE'],
        )

        self.assertEqual(t.state, TASK_STATES['FREE'])

        tm = TaskManager(conf={'worker': self._worker})
        task_info = t.export(False)

        with patch('kobo.worker.taskmanager.os', fork=Mock(return_value=9999)) as os_mock:
            tm.take_task(task_info)
            os_mock.fork.assert_called_once()

        self.assertEqual(tm.worker_info['current_load'], 100)
        self.assertFalse(tm.worker_info['ready'])

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['OPEN'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_fork_task(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForkTask',
            state=TASK_STATES['FREE'],
        )

        self.assertEqual(t.state, TASK_STATES['FREE'])

        tm = TaskManager(conf={'worker': self._worker})
        task_info = t.export(False)

        with patch('kobo.worker.taskmanager.os', fork=Mock(return_value=9999)) as os_mock:
            tm.fork_task(task_info)
            os_mock.fork.assert_called_once()

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_fork_task_runs_task_if_cant_fork(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForegroundTask',
            state=TASK_STATES['OPEN'],
        )

        self.assertEqual(t.state, TASK_STATES['OPEN'])

        tm = TaskManager(conf={'worker': self._worker})
        task_info = t.export(False)

        with patch('kobo.worker.taskmanager.os', fork=Mock(return_value=0)) as os_mock:
            os_mock.devnull = os.devnull

            with patch('kobo.worker.taskmanager.signal') as signal_mock:
                tm.fork_task(task_info)
                os_mock.fork.assert_called_once()
                os_mock.setpgrp.assert_called_once()
                os_mock._exit.assert_called_once()
                signal_mock.signal.assert_called()

        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['CLOSED'])

    def test_fork_task_logs_exceptions(self):
        """Exceptions from the child within fork_task are logged."""

        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForegroundTask',
            state=TASK_STATES['OPEN'],
        )

        logger = Mock()

        with patch('kobo.worker.taskmanager.HubProxy') as hub_mock:
            # Arrange for close_task call to fail (at end of task)
            hub_mock.return_value.worker.close_task.side_effect = RuntimeError("simulated error")

            tm = TaskManager(conf={'worker': self._worker}, logger=logger)
            task_info = t.export(False)

            with patch('kobo.worker.taskmanager.os', fork=Mock(return_value=0)) as os_mock:
                os_mock.devnull = os.devnull
                tm.fork_task(task_info)

        # It should have logged something about the failure to close the task.
        logger.log.assert_called_with(logging.CRITICAL, 'Error running forked task', exc_info=1)

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_run_task_runs_foreground_task(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForegroundTask',
            state=TASK_STATES['OPEN'],
        )

        self.assertEqual(t.state, TASK_STATES['OPEN'])

        tm = TaskManager(conf={'worker': self._worker})
        task_info = t.export(False)
 
        tm.run_task(task_info)

        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['CLOSED'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_run_task_runs_fork_task(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForkTask',
            state=TASK_STATES['OPEN'],
        )

        self.assertEqual(t.state, TASK_STATES['OPEN'])

        tm = TaskManager(conf={'worker': self._worker})
        task_info = t.export(False)

        tm.run_task(task_info)

        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['CLOSED'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_run_task_mark_task_as_failed(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyFailTask',
            state=TASK_STATES['OPEN'],
        )

        self.assertEqual(t.state, TASK_STATES['OPEN'])

        tm = TaskManager(conf={'worker': self._worker})
        task_info = t.export(False)

        tm.run_task(task_info)

        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['FAILED'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_run_task_mark_task_as_failed_if_generic_exception(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyExceptionTask',
            state=TASK_STATES['OPEN'],
        )

        self.assertEqual(t.state, TASK_STATES['OPEN'])

        tm = TaskManager(conf={'worker': self._worker})
        task_info = t.export(False)

        tm.run_task(task_info)

        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['FAILED'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_finish_task(self):
        tm = TaskManager(conf={'worker': self._worker})

        mock = Mock()
        # inject mock plugin
        tm.task_container.plugins['Mock'] = mock

        tm.finish_task({
            'method': 'Mock'
        })

        mock.cleanup.assert_called_once()
        mock.notification.assert_called_once()

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_is_finished_task(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForkTask',
            state=TASK_STATES['FREE'],
        )

        tm = TaskManager(conf={'worker': self._worker})
        task_info = t.export(False)

        with patch('kobo.worker.taskmanager.os', fork=Mock(return_value=9999)) as os_mock:
            tm.take_task(task_info)
            os_mock.fork.assert_called_once()

        with patch('kobo.worker.taskmanager.os', waitpid=Mock(return_value=(123, 0))) as os_mock:
            self.assertTrue(tm.is_finished_task(t.id))
            os_mock.waitpid.assert_called_once()

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_is_finished_task_invalid_child_pid(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForkTask',
            state=TASK_STATES['FREE'],
        )

        tm = TaskManager(conf={'worker': self._worker})
        task_info = t.export(False)

        with patch('kobo.worker.taskmanager.os', fork=Mock(return_value=9999)) as os_mock:
            tm.take_task(task_info)
            os_mock.fork.assert_called_once()

        with patch('kobo.worker.taskmanager.os', waitpid=Mock(return_value=(0, 0))) as os_mock:
            self.assertFalse(tm.is_finished_task(t.id))
            os_mock.waitpid.assert_called_once()

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_is_finished_task_catch_os_error(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForkTask',
            state=TASK_STATES['FREE'],
        )

        tm = TaskManager(conf={'worker': self._worker})
        task_info = t.export(False)

        with patch('kobo.worker.taskmanager.os', fork=Mock(return_value=9999)) as os_mock:
            tm.take_task(task_info)
            os_mock.fork.assert_called_once()

        err = OSError()
        err.errno = errno.ECHILD

        with patch('kobo.worker.taskmanager.os', waitpid=Mock(side_effect=err)) as os_mock:
            self.assertFalse(tm.is_finished_task(t.id))
            os_mock.waitpid.assert_called_once()

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_shutdown_with_running_tasks(self):
        t = Task.objects.create(
            worker=self._worker.worker,
            arch=self._arch,
            channel=self._channel,
            owner=self._user,
            method='DummyForkTask',
            state=TASK_STATES['FREE'],
        )

        self.assertEqual(t.state, TASK_STATES['FREE'])

        tm = TaskManager(conf={'worker': self._worker})
        task_info = t.export(False)

        with patch('kobo.worker.taskmanager.os', fork=Mock(return_value=9999)) as os_mock:
            tm.take_task(task_info)
            os_mock.fork.assert_called_once()

        tm.update_tasks()

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['OPEN'])

        tm.shutdown()

        # reload task info
        t = Task.objects.get(id=t.id)
        self.assertEqual(t.state, TASK_STATES['INTERRUPTED'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_shutdown_without_running_tasks(self):
        tm = TaskManager(conf={'worker': self._worker})
        tm.update_tasks()
        tm.shutdown()

        self.assertTrue(tm.worker_info['enabled'])
        self.assertTrue(tm.worker_info['ready'])

    @patch('kobo.worker.taskmanager.HubProxy', HubProxyMock)
    def test_lock(self):
        tm = TaskManager(conf={'worker': self._worker})
        self.assertFalse(tm.locked)
        tm.lock()
        self.assertTrue(tm.locked)



# -*- coding: utf-8 -*-

import unittest2 as unittest

from kobo.hub.xmlrpc.system import getAPIVersion


class TestApiVersion(unittest.TestCase):

    def test_api_version(self):
        version = getAPIVersion(None)
        self.assertEquals(version, '0.1.0')



# -*- coding: utf-8 -*-


from __future__ import print_function
from kobo.client import ClientCommand


class List_Workers(ClientCommand):
    """list workers"""
    enabled = True


    def options(self):
        self.parser.usage = "%%prog %s" % self.normalized_name

        self.parser.add_option(
            "--show-disabled",
            default=False,
            action="store_true",
            help="show disabled workers"
        )


    def run(self, *args, **kwargs):
        username = kwargs.pop("username", None)
        password = kwargs.pop("password", None)
        show_disabled = kwargs.get("show_disabled", False)

        self.set_hub(username, password)
        print(self.hub.client.list_workers(not show_disabled))



# -*- coding: utf-8 -*-

import threading
import time

import six

from six.moves import queue
from six.moves.xmlrpc_client import Fault
from six import BytesIO


__all__ = (
    "LoggingThread",
    "LoggingIO",
)


class LoggingThread(threading.Thread):
    """Send stdout data to hub in a background thread."""

    def __init__(self, hub, task_id, *args, **kwargs):
        threading.Thread.__init__(self, *args, **kwargs)
        self._hub = hub
        self._task_id = task_id
        self._queue = queue.Queue()
        self._event = threading.Event()
        self._running = True
        self._send_time = 0
        self._send_data = ""

    def run(self):
        """Send queue content to hub."""
        while self._running or not self._queue.empty() or self._send_data:
            if self._queue.empty():
                self._event.wait(5)

            self._event.clear()
            while True:
                try:
                    self._send_data += self._queue.get_nowait()
                except queue.Empty:
                    break

            if not self._send_data:
                continue

            now = int(time.time())
            if self._running and len(self._send_data) < 1200 and now - self._send_time < 5:
                continue

            if isinstance(self._send_data, six.text_type):
                self._send_data = self._send_data.encode('utf-8')

            try:
                self._hub.upload_task_log(BytesIO(self._send_data), self._task_id, "stdout.log", append=True)
                self._send_time = now
                self._send_data = ""
            except Fault:
                continue

    def write(self, data):
        """Add data to the queue and set the event for sending queue content."""
        self._queue.put(data)
        self._event.set()

    def stop(self):
        """Send remaining data to hub and finish."""
        self._running = False
        self._event.set()
        self.join()


class LoggingIO(object):
    """StringIO wrapper that also writes all data to a logging thread."""

    def __init__(self, io, logging_thread):
        self._io = io
        self._thread = logging_thread

    def __getattr__(self, name):
        return getattr(self._io, name)

    def write(self, data):
        """Write data to the IO stream and to the logging thread."""
        self._io.write(data)
        self._thread.write(data)



#!/usr/bin/python
# -*- coding: utf-8 -*-


import unittest

import tempfile
import os

from kobo.http import POSTTransport


class TestPOSTTransport(unittest.TestCase):
    def setUp(self):
        self.postt = POSTTransport()

    def test_get_content_type(self):
        tf0 = tempfile.mkstemp()[1]
        tf1 = tempfile.mkstemp(suffix=".txt")[1]
        tf2 = tempfile.mkstemp(suffix=".rtf")[1]
        tf3 = tempfile.mkstemp(suffix=".avi")[1]
        self.assertEqual(self.postt.get_content_type(tf0), "application/octet-stream")
        self.assertEqual(self.postt.get_content_type(tf1), "text/plain")
        # *.rtf: py2.7 returns 'application/rtf'; py2.4 returns 'text/rtf'
        self.assertEqual(self.postt.get_content_type(tf2).split("/")[1], "rtf")
        self.assertTrue(self.postt.get_content_type(tf2) in ("application/rtf", "text/rtf"))
        self.assertEqual(self.postt.get_content_type(tf3), "video/x-msvideo")

    def test_add_file(self):
        tf1 = tempfile.mkstemp()[1]
        tf2 = tempfile.mkstemp()[1]
        tf3 = open(tempfile.mkstemp()[1])
        os.unlink(tf1)
        self.assertRaises(OSError, self.postt.add_file, "file", tf1)
        self.assertEqual(self.postt.add_file("file", tf2), None)
        self.assertRaises(TypeError, self.postt.add_file, "file", tf3)



# -*- coding: utf-8 -*-

from kobo.django.fields import JSONField

from django.db import models


class DummyModel(models.Model):
    field = JSONField()


class DummyDefaultModel(models.Model):
    field = JSONField(default={})

class DummyNotHumanModel(models.Model):
    field = JSONField(default={}, human_readable = True)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('hub', '0002_auto_20150722_0612'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='task',
            options={'ordering': ('-id',), 'permissions': (('can_see_traceback', 'Can see traceback'),)},
        ),
        migrations.AddField(
            model_name='worker',
            name='min_priority',
            field=models.PositiveIntegerField(default=0, help_text='Worker will take only tasks of this or higher priority.'),
        ),
    ]
